#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test038_gan_phsp_spect_gan_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test038_gan_phsp_spect")
paths.output_ref = paths.output_ref / "test038_ref"

# create the simulation
sim = gate.Simulation()
condition_generator = create_simulation(sim, paths)

# go
sim.initialize()
sim.start()

# test
all_cond = condition_generator.all_cond
analyze_results(sim, paths, all_cond)



import opengate as gate
import copy

element_builders = {
    "Volume": gate.volume_builders,
    "Source": gate.source_builders,
    "Actor": gate.actor_builders,
    "Filter": gate.filter_builders,
}


def get_element_class(element_type, type_name):
    """
    Return the class of the given type_name (in the element_type list)
    """
    elements = None
    if element_type == "Volume":
        elements = gate.volume_type_names
    if element_type == "Source":
        elements = gate.source_type_names
    if element_type == "Actor":
        elements = gate.actor_type_names
    if element_type == "Filter":
        elements = gate.filter_type_names
    if not elements:
        gate.fatal(
            f"Error, element_type={element_type} is   unknown. Use Volume, Source or Actor."
        )
    for e in elements:
        # check the class has type_name
        if not hasattr(e, "type_name"):
            gate.fatal(
                f'Error, the class {e.__name__} *must* have a static attribute called "type_name"'
            )
        # is the type the one we are looking ?
        if e.type_name == type_name:
            return e
    s = [x.type_name for x in elements]
    gate.fatal(
        f'Error {element_type}: the type "{type_name}" is unknown. Known types are {s}'
    )


def get_builder(element_type, type_name):
    """
    Return a function that build an element of the class type_name
    Check everything first.
    """
    # get type of element builder
    if element_type not in element_builders:
        gate.fatal(
            f"The element type: {element_type} is unknown.\n"
            f"Known element types are {element_builders.keys()}"
        )
    builders = element_builders[element_type]
    # get builder
    if type_name not in builders:
        gate.fatal(
            f"The element type: {type_name} is unknown.\n"
            f"Known type names are {builders.keys()}"
        )
    builder = builders[type_name]
    return builder


def new_element(user_info, simulation=None):
    """
    Create a new element (Volume, Source, Actor, Filter), according to the type name
    - use the element_builders to find the class to build
    - create a new element, with the name as parameter to the constructor
    - initialize the default list of keys in the user_info
    - set a pointer to the Simulation object
    """
    builder = get_builder(user_info.element_type, user_info.type_name)
    # build (create the object)
    e = builder(user_info)
    # set the simulation pointer
    e.set_simulation(simulation)
    return e


def copy_user_info(v1, v2):
    """
    Copy all attributes from v1 to v2, except the name.
    v1 is assumed to be a UserInfo object with several attribute members.
    v2 must have the (at least) the same set of attributes.
    Values are (deep) copied.
    """
    for k in v1.__dict__:
        if k == "name":
            continue
        if k == "_name":
            continue
        setattr(v2, k, copy.deepcopy(v1.__dict__[k]))



import numpy as np
import os
import opengate as gate
from box import Box
import gatetools.phsp as phsp
import uproot
import matplotlib.pyplot as plt


def root_compare_param_tree(filename, tree_name, keys):
    p = Box()
    p.root_file = filename
    p.tree_name = tree_name
    p.the_keys = keys
    p.scaling = [1.0] * len(keys)
    p.mins = [None] * len(keys)
    p.maxs = [None] * len(keys)
    return p


def root_compare_param(keys, fig):
    p = Box()
    p.tols = [1.0] * len(keys)
    p.fig = fig
    p.hits_tol = 6
    p.nb_bins = 300
    return p


def root_compare4(p1, p2, param):
    """
    Compare two root trees.

    p1 and p2 contain = the root filename, the tree name, a list of branch names
    (see root_compare_param_tree)
    Also, each branch can be scaled and clip to a min/max range.

    param contains: the tolerance values (for all branches), the fig name,
    the nb of bins of the histograms, the tolerance for the nb of hits
    (see root_compare_param)
    """

    # read reference tree
    hits1 = uproot.open(p1.root_file)[p1.tree_name]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    # read tree
    hits2 = uproot.open(p2.root_file)[p2.tree_name]
    hits2_n = hits2.num_entries
    hits2 = hits2.arrays(library="numpy")

    print(f"Reference tree: {os.path.basename(p1.root_file)} n={hits1_n}")
    print(f"Current tree:   {os.path.basename(p2.root_file)} n={hits2_n}")
    diff = gate.rel_diff(float(hits2_n), float(hits1_n))
    b = np.fabs(diff) < param.hits_tol
    is_ok = gate.print_test(b, f"Difference: {hits1_n} {hits2_n} {diff:+.2f}%")
    print(f"Reference tree: {hits1.keys()}")
    print(f"Current tree:   {hits2.keys()}")

    # compare the given tree
    p1.tree = hits1
    p2.tree = hits2
    is_ok = compare_trees4(p1, p2, param) and is_ok

    # figure
    plt.suptitle(
        f"Values: ref {os.path.basename(p1.root_file)} {os.path.basename(p2.root_file)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(param.fig)
    print(f"Figure in {param.fig}")

    return is_ok


def compare_trees4(p1, p2, param):
    nb_fig = 0
    ax = None
    if param.fig:
        nb_fig = len(p1.the_keys)
        nrow, ncol = phsp.fig_get_nb_row_col(nb_fig)
        f, ax = plt.subplots(nrow, ncol, figsize=(25, 10))
    is_ok = True
    n = 0
    print("Compare branches with Wasserstein distance")
    for i in range(len(p2.the_keys)):
        if param.fig:
            a = phsp.fig_get_sub_fig(ax, i)
            n += 1
        else:
            a = False
        b1 = p1.tree[p1.the_keys[i]] * p1.scaling[i]
        b2 = p2.tree[p2.the_keys[i]] * p2.scaling[i]
        if p1.mins[i] is not None:
            b1 = b1[b1 > p1.mins[i]]
        if p1.maxs[i] is not None:
            b1 = b1[b1 < p1.maxs[i]]
        if p2.mins[i] is not None:
            b2 = b2[b2 > p2.mins[i]]
        if p2.maxs[i] is not None:
            b2 = b2[b2 < p2.maxs[i]]
        is_ok = (
            gate.compare_branches_values(
                b1, b2, p1.the_keys[i], p2.the_keys[i], param.tols[i], a, param.nb_bins
            )
            and is_ok
        )

    if param.fig:
        phsp.fig_rm_empty_plot(nb_fig, n, ax)
    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.pet_philips_vereos as pet_vereos
import opengate.contrib.phantom_necr as phantom_necr

paths = gate.get_default_test_paths(__file__, "gate_test037_pet")

"""
Simulation of a PET VEREOS with NEMA NECR phantom.
- phantom is a simple cylinder and linear source
- output is hits and singles only (no coincidences)
- also digitizer is simplified: only raw hits and adder (for singles)
"""

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.check_volumes_overlap = False
ui.g4_verbose = True
ui.g4_verbose_level = 2
# ui.number_of_threads = 2
# ui.force_multithread_mode = True

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
Bq = gate.g4_units("Bq")
MBq = Bq * 1e6
sec = gate.g4_units("second")

#  change world size
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]
world.material = "G4_AIR"

# add a PET VEREOS
sim.add_material_database(paths.gate_data / "GateMaterials_pet.db")
pet = pet_vereos.add_pet_debug(sim, "pet", create_housing=False, create_mat=False)

# add table
# bed = pet_vereos.add_table(sim, "pet")

# add NECR phantom
# phantom = phantom_necr.add_necr_phantom(sim, "phantom")

# physics
p = sim.get_physics_user_info()
# p.enable_decay = True #not needed (?)
p.physics_list_name = "G4EmStandardPhysics_option4"
sim.set_cut("world", "all", 1 * m)
sim.set_cut(f"{pet.name}_crystal", "all", 1 * mm)

# em = p.g4_em_parameters
# print('em', em)
import opengate_core as g4

em = g4.G4EmParameters.Instance()
print("em", em)
em.SetFluo(True)
em.SetAuger(True)
em.SetAugerCascade(True)
em.SetPixe(True)
em.SetApplyCuts(True)

print("em pix", em.Pixe())
print("em fluo", em.Fluo())
print("em Auger", em.Auger())
print("em AugerCascade", em.AugerCascade())
print("em ApplyCuts", em.ApplyCuts())
print("em min E", em.MinKinEnergy())
print("em max E", em.MaxKinEnergy())

# emPar->SetMinEnergy(mEmin); ???


# default source for tests
source = phantom_necr.add_necr_source_debug(sim, "mysource")
total_yield = gate.get_rad_yield("F18")
print("Yield for F18 (nb of e+ per decay) : ", total_yield)
source.activity = 1787.914158 * MBq * total_yield
source.half_life = 6586.26 * sec
# source.energy.type = "F18_analytic"  # WARNING not ok, but similar to previous Gate
# source.energy.type = "F18"  # this is the correct F18 e+ source

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# hits collection
hc = sim.add_actor("HitsCollectionActor", "Hits")
# get crystal volume by looking for the word crystal in the name
l = sim.get_all_volumes_user_info()
crystal = l[[k for k in l if "crystal" in k][0]]
hc.mother = crystal.name
print("Crystal :", crystal.name)
hc.output = paths.output / "test037_test0.root"
hc.attributes = [
    "PostPosition",
    "TotalEnergyDeposit",
    "TrackVolumeCopyNo",
    "PreStepUniqueVolumeID",
    "PostStepUniqueVolumeID",
    "GlobalTime",
    # "KineticEnergy", # not needed
    # "ProcessDefinedStep", # not needed
]
hc.keep_zero_edep = False  # 'False' is the default
hc.debug = False

# singles collection
sc = sim.add_actor("HitsAdderActor", "Singles")
sc.mother = crystal.name
sc.input_hits_collection = "Hits"
# sc.policy = "EnergyWinnerPosition"
sc.policy = "EnergyWeightedCentroidPosition"
# the following attributes is not needed in singles
# sc.skip_attributes = ["KineticEnergy"]
sc.output = hc.output

# timing
sim.run_timing_intervals = [[0, 0.0001 * sec]]
if ui.visu:
    sim.run_timing_intervals = [[0, 0.00000002 * sec]]

eV = gate.g4_units("eV")
GeV = gate.g4_units("GeV")

# pct = g4.G4ProductionCutsTable.GetProductionCutsTable()
# pct.SetEnergyRange(250 * eV, 0.5 * GeV) ## do nothing :(

# create G4 objects
sim.initialize()
# sim.check_volumes_overlap(True)

em = g4.G4EmParameters.Instance()
em.SetFluo(True)
em.SetAuger(True)
em.SetAugerCascade(True)
em.SetPixe(True)
em.SetApplyCuts(True)
print("2 em", em)
print("em pix", em.Pixe())
print("em fluo", em.Fluo())
print("em Auger", em.Auger())
print("em AugerCascade", em.AugerCascade())
print("em fluo", em.ApplyCuts())
print("em min E", em.MinKinEnergy())
print("em max E", em.MaxKinEnergy())

# em.SetMinEnergy(250 * eV)
# print('em min E', em.MinKinEnergy())
# print('em max E', em.MaxKinEnergy())
# p.enable_decay = True #not needed (?)

# pct = g4.G4ProductionCutsTable.GetProductionCutsTable()
# pct.SetEnergyRange(250 * eV, 0.5 * GeV)

# sim.set_cut("world", "all", 1 * m)

vm = sim.volume_manager
t = vm.dump_defined_material(10)
print(t)
# exit(0)

m = vm.find_or_build_material("LYSO_debug")
print(m)
# exit(0)

# start simulation
sim.start()

em = g4.G4EmParameters.Instance()
print("3em", em)
print("em pix", em.Pixe())
print("em fluo", em.Fluo())
print("em Auger", em.Auger())
print("em AugerCascade", em.AugerCascade())
print("em fluo", em.ApplyCuts())

# print results
stats = sim.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
p = paths.gate / "output"
stats_ref = gate.read_stat_file(p / "stats0.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

print()
gate.warning(f"Check stats")
p = paths.gate / "output_test1"
stats_ref = gate.read_stat_file(p / "stats1.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check phsp (new version)
p = paths.gate / "output"
k1 = ["posX", "posY", "posZ", "edep", "time"]
k2 = [
    "PostPosition_X",
    "PostPosition_Y",
    "PostPosition_Z",
    "TotalEnergyDeposit",
    "GlobalTime",
]
p1 = gate.root_compare_param_tree(p / "output0.root", "Hits", k1)
p2 = gate.root_compare_param_tree(hc.output, "Hits", k2)
p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
p = gate.root_compare_param(p1.the_keys, paths.output / "test037_test0_hits.png")
p.tols[k1.index("posX")] = 3
p.tols[k1.index("posY")] = 3.5
p.tols[k1.index("posZ")] = 0.5
p.tols[k1.index("edep")] = 0.002
p.tols[k1.index("time")] = 0.0001
is_ok = gate.root_compare4(p1, p2, p) and is_ok

print()
print()
print()

# check phsp (hits)
print()
gate.warning(f"Check root")
p = paths.gate / "output"
ref_file = p / "output0.root"
hc_file = hc.output
checked_keys = [
    "PostPosition_X",
    "PostPosition_Y",
    "PostPosition_Z",
    "TotalEnergyDeposit",
    "GlobalTime",
]
checked_keys_ref = ["posX", "posY", "posZ", "edep", "time"]
scalings_ref = [1.0] * len(checked_keys)
scalings = [1.0] * len(checked_keys)
scalings[checked_keys.index("GlobalTime")] = 1e-9  # time in ns
tols = [1.0] * len(checked_keys)
tols[checked_keys_ref.index("posX")] = 3
tols[checked_keys_ref.index("posY")] = 3.5
tols[checked_keys_ref.index("posZ")] = 0.5
tols[checked_keys_ref.index("edep")] = 0.002
tols[checked_keys_ref.index("time")] = 0.0001

is_ok = (
    gate.compare_root3(
        ref_file,
        hc_file,
        "Hits",
        "Hits",
        checked_keys_ref,
        checked_keys,
        tols,
        scalings_ref,
        scalings,
        paths.output / "test037_test0_hits.png",
        nb_bins=300,
    )
    and is_ok
)

print()
import uproot

hits1 = uproot.open(ref_file)["Hits"].arrays(library="numpy")
hits2 = uproot.open(hc_file)["Hits"].arrays(library="numpy")
edep1 = hits1["edep"]
edep2 = hits2["TotalEnergyDeposit"]
print(edep1.shape, edep2.shape)
epsilon = 0
a1 = edep1[edep1 <= epsilon]
a2 = edep2[edep2 <= epsilon]
print(a1, a1.shape)
print(a2, a2.shape)
print(f"Nb of zero in edep1 {len(a1)}")
print(f"Nb of zero in edep2 {len(a2)}")

print()

# check phsp (singles)
print()
gate.warning(f"Check root")
checked_keys_ref = ["globalPosX", "globalPosY", "globalPosZ", "energy", "time"]
tols[checked_keys_ref.index("energy")] = 0.02
is_ok = (
    gate.compare_root3(
        ref_file,
        hc_file,
        "Singles",
        "Singles",
        checked_keys_ref,
        checked_keys,
        tols,
        scalings_ref,
        scalings,
        paths.output / "test037_test0_singles.png",
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test013_phys_lists_base import create_pl_sim, phys_em_parameters
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# create simulation
sim = create_pl_sim()

# keep only ion sources
# sim.source_manager.user_info_sources.pop("gamma")

# change physics
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
p.enable_decay = True
p.apply_cuts = True  # default
cuts = p.production_cuts
mm = gate.g4_units("mm")
cuts.world.gamma = 1 * mm
cuts.world.electron = 0.1 * mm
cuts.world.positron = 1 * mm
cuts.world.proton = 1 * mm
cuts.b1.gamma = 1 * mm
cuts.b1.electron = 0.01 * mm
cuts.b1.positron = 1 * mm
cuts.b1.proton = 1 * mm

# initialize
sim.initialize()

# em parameters
phys_em_parameters(p)

print("Phys list cuts:")
print(sim.physics_manager.dump_cuts())

# start simulation
# sim.set_g4_verbose(True)
# sim.apply_g4_command("/tracking/verbose 1")
sim.start()

# Gate mac/main_3.mac
stats = sim.get_actor("Stats")
stats_ref = gate.read_stat_file(
    pathFile
    / ".."
    / "data"
    / "gate"
    / "gate_test013_phys_lists"
    / "output"
    / "stat_3.txt"
)
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.1)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4


class TrapVolume(gate.VolumeBase):
    """
    http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html
    dx1 Half x length of the side at y=-pdy1 of the face at -pdz
    dx2 Half x length of the side at y=+pdy1 of the face at -pdz
    dz Half z length
    theta Polar angle of the line joining the centres of the faces at -/+pdz
    phi Azimuthal angle of the line joining the centre of the face at -pdz to the centre of the face at +pdz
    dy1 Half y length at -pdz
    dy2 Half y length at +pdz
    dx3 Half x length of the side at y=-pdy2 of the face at +pdz
    dx4 Half x length of the side at y=+pdy2 of the face at +pdz
    alp1 Angle with respect to the y axis from the centre of the side (lower endcap)
    alp2 Angle with respect to the y axis from the centre of the side (upper endcap)
    """

    type_name = "Trap"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        u = user_info
        mm = gate.g4_units("mm")
        u.dx1 = 30 * mm
        u.dx2 = 40 * mm
        u.dy1 = 40 * mm
        u.dx3 = 10 * mm
        u.dx4 = 14 * mm
        u.dy2 = 16 * mm
        u.dz = 60 * mm
        deg = gate.g4_units("deg")
        u.theta = 20 * deg
        u.phi = 5 * deg
        u.alp1 = u.alp2 = 10 * deg

    def build_solid(self):
        u = self.user_info
        return g4.G4Trap(
            u.name,
            u.dz,
            u.theta,
            u.phi,
            u.dy1,
            u.dx1,
            u.dx2,
            u.alp1,
            u.dy2,
            u.dx3,
            u.dx4,
            u.alp2,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
import pathlib
import os

pathFile = pathlib.Path(__file__).parent.resolve()

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
nm = gate.g4_units("nm")
mm = gate.g4_units("mm")

#  change world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [10 * cm, 10 * cm, 10 * cm]
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]

# default source for tests
source = sim.add_source("Generic", "mysource")
source.energy.mono = 50 * MeV
source.particle = "proton"
source.position.type = "sphere"
source.position.radius = 1 * cm
source.direction.type = "iso"
source.activity = 10000 * Bq

# filter : keep gamma
f = sim.add_filter("ParticleFilter", "f")
f.particle = "gamma"
fp = sim.add_filter("ParticleFilter", "fp")
fp.particle = "e-"

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = pathFile / ".." / "output" / "test023-edep.mhd"
# dose.output = 'output_ref/test023-edep.mhd'
dose.mother = "waterbox"
dose.size = [100, 100, 100]
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.filters.append(fp)

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True
s.filters.append(f)

print(s)
print(dose)
print("Filters: ", sim.filter_manager)
print(sim.filter_manager.dump())

# change physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMZ"
cuts = p.production_cuts
cuts.world.gamma = 0.1 * mm
cuts.world.proton = 0.1 * mm
cuts.world.electron = 0.1 * mm
cuts.world.positron = 0.1 * mm

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("Stats")
print(stat)
# stat.write('output_ref/test023_stats.txt')

# tests
stats_ref = gate.read_stat_file(
    pathFile / ".." / "data" / "output_ref" / "test023_stats.txt"
)
is_ok = gate.assert_stats(stat, stats_ref, 0.8)
is_ok = is_ok and gate.assert_images(
    pathFile / ".." / "data" / "output_ref" / "test023-edep.mhd",
    pathFile / ".." / "output" / "test023-edep.mhd",
    stat,
    tolerance=50,
)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4


class HitsAdderActor(g4.GateHitsAdderActor, gate.ActorBase):
    """
    Equivalent to Gate "adder": gather all hits of an event in the same volume.
    Input: a HitsCollection, need aat least TotalEnergyDeposit and PostPosition attributes
    Output: a Single collections

    Policies:
    - TakeEnergyWinner: consider position and energy of the hit with the max energy
       -> all other attributes (Time, etc): the value of the winner is used.
    - TakeEnergyCentroid: computed the energy-weighted centroid
       -> all other attributes (Time, etc): the value the last seen hit is used.

    """

    type_name = "HitsAdderActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.attributes = []
        user_info.output = "singles.root"
        user_info.input_hits_collection = "Hits"
        user_info.policy = "TakeEnergyWinner"
        user_info.skip_attributes = []
        user_info.clear_every = 1e5

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateHitsAdderActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)
        self.fStepFillNames = user_info.attributes

    def __del__(self):
        pass

    def __str__(self):
        s = f"HitsAdderActor {self.user_info.name}"
        return s

    def StartSimulationAction(
        self,
    ):  # not needed, only if need to do something in python
        g4.GateHitsAdderActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateHitsAdderActor.EndSimulationAction(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
from scipy.spatial.transform import Rotation
import pathlib
import os

pathFile = pathlib.Path(__file__).parent.resolve()

# global log level
# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
cm3 = gate.g4_units("cm3")
Bq = gate.g4_units("Bq")
BqmL = Bq / cm3
print("Unit Bq", Bq)
print("Unit mL (cm3)", cm3)
print("Unit BqmL", BqmL)

# change world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a iec phantom
iec_phantom = gate_iec.add_phantom(sim)
iec_phantom.translation = [3 * cm, 1 * cm, 0 * cm]
iec_phantom.rotation = Rotation.from_euler("y", 33, degrees=True).as_matrix()

# simple source
# gate_iec.add_sources(sim, 'iec', 'all')
ac = 2000 * BqmL
gate_iec.add_spheres_sources(
    sim,
    "iec",
    "iec_source",
    [10, 13, 17, 22, 28, 37],
    # [ac, 0, 0, 0, 0, 0])
    [ac, ac, ac, ac, ac, ac],
    verbose=True,
)

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "stats")
stats.track_types_flag = True

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = pathFile / ".." / "output" / "test015.mhd"
# dose.output = 'output_ref/test015_ref.mhd'
dose.mother = "iec"
dose.size = [100, 100, 100]
mm = gate.g4_units("mm")
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.translation = [0 * mm, 0 * mm, 0 * mm]

# run timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 1 * sec]]

print(sim.volume_manager.dump_tree())
print(sim.source_manager.dump())

# initialize & start
sim.initialize()
sim.start()

# Only for reference stats:
stats = sim.get_actor("stats")
# stats.write('output_ref/test015_stats.txt')

# check
stats_ref = gate.read_stat_file(
    pathFile / ".." / "data" / "output_ref" / "test015_stats.txt"
)
is_ok = gate.assert_stats(stats, stats_ref, 0.07)
is_ok = is_ok and gate.assert_images(
    pathFile / ".." / "data" / "output_ref" / "test015_ref.mhd",
    pathFile / ".." / "output" / "test015.mhd",
    stats,
    tolerance=65,
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import pathlib
import os

pathFile = pathlib.Path(__file__).parent.resolve()

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.check_volumes_overlap = False

#  change world size
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]

# add a box (not really useful here)
# prefer air to speed simulation
airbox = sim.add_volume("Box", "Airbox")
airbox.size = [30 * cm, 30 * cm, 30 * cm]
airbox.translation = [0 * cm, 0 * cm, 0 * cm]
airbox.material = "G4_AIR"
airbox.color = [0, 0, 1, 1]  # blue

# add a PET ... or two PET !
import opengate.contrib.pet_philips_vereos as gate_pet

pet1 = gate_pet.add_pet(sim, "pet1")
# pet2 = gate_vereos.add_pet(sim, 'pet2')
# pet2.translation = [0, 0, pet1.dz * 2]

# default source for tests
source = sim.add_source("Generic", "Default")
Bq = gate.g4_units("Bq")
source.particle = "e+"
source.energy.type = "F18"
source.position.type = "sphere"
source.position.radius = 5 * cm
source.position.translation = [0, 0, 0]
source.direction.type = "iso"
source.activity = 1000 * Bq

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# create G4 objects
sim.initialize()

# explicit check overlap (already performed during initialize)
# sim.check_volumes_overlap(verbose=True)

# start simulation
sim.start()

# print results
stats = sim.get_actor("Stats")
# stats.write('output_ref/test018_stats_ref.txt')

# check
stats = sim.get_actor("Stats")
stats_ref = gate.read_stat_file(
    pathFile / ".." / "data" / "output_ref" / "test018_stats_ref.txt"
)
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.15)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("second")
deg = gate.g4_units("deg")
kBq = 1000 * Bq
MBq = 1000 * kBq

""" ================================================== """
# main parameters
ui = sim.user_info
ui.g4_verbose = False
ui.check_volumes_overlap = False
ui.number_of_threads = 1
ac = 10 * MBq
ac = 1 * kBq

# visu
ui.visu = False
ui.visu_verbose = False
colli_flag = not ui.visu
if ui.visu:
    ac = 100 * Bq
    ui.number_of_threads = 1

# colli param
distance = 15 * cm
psd = 6.11 * cm
p = [0, 0, -(distance + psd)]
""" ================================================== """

# world size
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]
world.material = "G4_AIR"

# spect head (debug mode = very small collimator)
spect1 = gate_spect.add_ge_nm67_spect_head(
    sim, "spect1", collimator_type=colli_flag, debug=False
)
spect1.translation, spect1.rotation = gate.get_transform_orbiting(p, "x", 180)

# spect head (debug mode = very small collimator)
spect2 = gate_spect.add_ge_nm67_spect_head(
    sim, "spect2", collimator_type=colli_flag, debug=False
)
spect2.translation, spect2.rotation = gate.get_transform_orbiting(p, "x", 0)

# physic list
sim.set_cut("world", "all", 10 * mm)
# sim.set_cut('spect1_crystal', 'all', 1 * mm)
# sim.set_cut('spect2_crystal', 'all', 1 * mm)

# source #1
sources = []
source = sim.add_source("Generic", "source1")
source.particle = "gamma"
source.energy.type = "mono"
source.energy.mono = 140.5 * keV
source.position.type = "sphere"
source.position.radius = 2 * mm
source.position.translation = [0, 0, 20 * mm]
source.direction.type = "iso"
source.direction.acceptance_angle.volumes = ["spect2", "spect1"]
source.direction.acceptance_angle.intersection_flag = True
source.direction.acceptance_angle.normal_flag = True
source.direction.acceptance_angle.normal_vector = [0, 0, -1]
source.direction.acceptance_angle.normal_tolerance = 10 * deg
source.activity = ac / ui.number_of_threads
sources.append(source)

# source #1
source2 = sim.add_source("Generic", "source2")
gate.copy_user_info(source, source2)
source2.position.radius = 1 * mm
source2.position.translation = [20 * mm, 0, -20 * mm]
sources.append(source2)

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.output = paths.output / "test033_stats.txt"

# add default digitizer (it is easy to change parameters if needed)
gate_spect.add_simplified_digitizer_Tc99m(
    sim, "spect1_crystal", paths.output / "test033_proj_1.mhd"
)
gate_spect.add_simplified_digitizer_Tc99m(
    sim, "spect2_crystal", paths.output / "test033_proj_2.mhd"
)

# motion of the spect, create also the run time interval
heads = [spect1, spect2]

# create a list of run (total = 1 second)
n = 10
sim.run_timing_intervals = gate.range_timing(0, 1 * sec, n)

for head in heads:
    motion = sim.add_actor("MotionVolumeActor", f"Move_{head.name}")
    motion.mother = head.name
    motion.translations, motion.rotations = gate.volume_orbiting_transform(
        "x", 0, 180, n, head.translation, head.rotation
    )
    motion.priority = 5

# go
sim.initialize()
sim.start()

stats = sim.get_actor("Stats")
print(stats)

s = 0
for source in sources:
    s += gate.get_source_skipped_particles(sim, source.name)
print(f"Skipped particles {s}")

########################
gate.warning(f"Check skipped")
ref_skipped = 19695798
tol = 0.05
d = abs(ref_skipped - s) / ref_skipped
is_ok = d < tol
gate.print_test(
    is_ok,
    f"Skipped particles ref={ref_skipped}, get {s} -> {d * 100}% vs tol={tol * 100}%",
)

########################
gate.warning(f"Check stats")
stats_ref = gate.read_stat_file(paths.output_ref / "test033_stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.05) and is_ok

# compare edep map
gate.warning(f"Check images")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test033_proj_1.mhd",
        paths.output / "test033_proj_1.mhd",
        stats,
        tolerance=45,
        axis="x",
    )
    and is_ok
)
is_ok = (
    gate.assert_images(
        paths.output_ref / "test033_proj_2.mhd",
        paths.output / "test033_proj_2.mhd",
        stats,
        tolerance=45,
        axis="x",
    )
    and is_ok
)

gate.test_ok(is_ok)



import opengate_core as g4


class RunAction(g4.G4UserRunAction):
    """
    User action at begin and end of run.
    Every time a run begin/end, the G4 engine call BeginOfRunAction and EndOfRunAction.
    The callback is then forwarded to all actors that need it.
    """

    def __init__(self):
        g4.G4UserRunAction.__init__(self)
        self.BeginOfRunAction_actors = []
        self.EndOfRunAction_actors = []

    def __del__(self):
        pass

    def register_actor(self, actor):
        actions = actor.actions
        if "BeginOfRunAction" in actions:
            self.BeginOfRunAction_actors.append(actor)
        if "EndOfRunAction" in actions:
            self.EndOfRunAction_actors.append(actor)

    def BeginOfRunAction(self, run):
        for actor in self.BeginOfRunAction_actors:
            actor.BeginOfRunAction(run)

    def EndOfRunAction(self, run):
        for actor in self.EndOfRunAction_actors:
            # actor.ProcessHitsPerBatch(True) # already done in GateVActor
            actor.EndOfRunAction(run)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
import opengate.contrib.phantom_nema_iec_body as gate_iec
from scipy.spatial.transform import Rotation
import numpy as np

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.check_volumes_overlap = False
ui.random_seed = 123456

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("second")
deg = gate.g4_units("deg")
kBq = 1000 * Bq
MBq = 1000 * kBq

""" ================================================== """
# main parameters
ui.visu = False
ui.g4_verbose = False
ui.visu_verbose = False
ui.number_of_threads = 2
"""
WARNING : it works but it is much slower than mono-thread simulation.
Indeed, between each run the master synchronize all workers and this requires that
all workers finish their jobs (hence longer one).
Once all threads are synchronized, the geometry is moved.
Then only, workers can start their next run.

It is unknown if this MT version can be useful (yet), maybe for very large
simulation ?
"""
ac = 10 * MBq
ac = 1 * MBq
distance = 15 * cm
psd = 6.11 * cm
p = [0, 0, -(distance + psd)]
""" ================================================== """

# world size
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]
world.material = "G4_AIR"

# spect head (debug mode = very small collimator)
spect1 = gate_spect.add_ge_nm67_spect_head(
    sim, "spect1", collimator_type="lehr", debug=ui.visu
)
spect1.translation, spect1.rotation = gate.get_transform_orbiting(p, "x", 180)

# spect head (debug mode = very small collimator)
spect2 = gate_spect.add_ge_nm67_spect_head(
    sim, "spect2", collimator_type="lehr", debug=ui.visu
)
spect2.translation, spect2.rotation = gate.get_transform_orbiting(p, "x", 0)

# physic list
sim.set_cut("world", "all", 10 * mm)
# sim.set_cut('spect1_crystal', 'all', 1 * mm)
# sim.set_cut('spect2_crystal', 'all', 1 * mm)

# source #1
sources = []
source = sim.add_source("Generic", "source1")
source.particle = "gamma"
source.energy.type = "mono"
source.energy.mono = 140.5 * keV
source.position.type = "sphere"
source.position.radius = 2 * mm
source.position.translation = [0, 0, 20 * mm]
source.direction.type = "iso"
source.direction.acceptance_angle.volumes = ["spect2", "spect1"]
source.direction.acceptance_angle.intersection_flag = True
source.direction.acceptance_angle.normal_flag = True
source.direction.acceptance_angle.normal_vector = [0, 0, -1]
source.direction.acceptance_angle.normal_tolerance = 10 * deg
source.activity = ac / ui.number_of_threads
sources.append(source)

# source #1
source2 = sim.add_source("Generic", "source2")
gate.copy_user_info(source, source2)
source2.position.radius = 1 * mm
source2.position.translation = [20 * mm, 0, -20 * mm]
sources.append(source2)

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.output = paths.output / "test033_stats.txt"

# add default digitizer (it is easy to change parameters if needed)
proj = gate_spect.add_simplified_digitizer_Tc99m(
    sim, "spect1_crystal", paths.output / "test033_proj_1.mhd"
)
proj.origin_as_image_center = False
proj = gate_spect.add_simplified_digitizer_Tc99m(
    sim, "spect2_crystal", paths.output / "test033_proj_2.mhd"
)
proj.origin_as_image_center = False

# motion of the spect, create also the run time interval
heads = [spect1, spect2]

# create a list of run (total = 1 second)
n = 10
sim.run_timing_intervals = gate.range_timing(0, 1 * sec, n)

for head in heads:
    motion = sim.add_actor("MotionVolumeActor", f"Move_{head.name}")
    motion.mother = head.name
    motion.translations, motion.rotations = gate.volume_orbiting_transform(
        "x", 0, 180, n, head.translation, head.rotation
    )
    motion.priority = 5

# go
sim.initialize()
sim.start()

stats = sim.get_actor("Stats")
print(stats)

s = 0
for source in sources:
    s += gate.get_source_skipped_particles(sim, source.name)
print(f"Skipped particles {s}")

########################
gate.warning(f"Check skipped")
ref_skipped = 1970136
tol = 0.05
d = abs(ref_skipped - s) / ref_skipped
is_ok = d < tol
gate.print_test(
    is_ok,
    f"Skipped particles ref={ref_skipped}, get {s} -> {d * 100}% vs tol={tol * 100}%",
)

########################
gate.warning(f"Check stats")
stats_ref = gate.read_stat_file(paths.output_ref / "test033_stats.txt")
stats_ref.counts.run_count *= ui.number_of_threads
is_ok = gate.assert_stats(stats, stats_ref, 0.05) and is_ok

# compare edep map
gate.warning(f"Check images")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test033_proj_1.mhd",
        paths.output / "test033_proj_1.mhd",
        stats,
        tolerance=71,
        axis="x",
    )
    and is_ok
)
is_ok = (
    gate.assert_images(
        paths.output_ref / "test033_proj_2.mhd",
        paths.output / "test033_proj_2.mhd",
        stats,
        tolerance=80,
        axis="x",
    )
    and is_ok
)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4
from box import Box
from scipy.spatial.transform import Rotation

rid = Rotation.identity().as_matrix()

bool_operators = ["union", "subtraction", "intersection"]


def solid_union(a, b, tr=None, rot=rid):
    if tr is None:
        tr = [0, 0, 0]
    return solid_bool("union", a, b, tr, rot)


def solid_subtraction(a, b, tr=None, rot=rid):
    if tr is None:
        tr = [0, 0, 0]
    return solid_bool("subtraction", a, b, tr, rot)


def solid_intersection(a, b, tr=None, rot=rid):
    if tr is None:
        tr = [0, 0, 0]
    return solid_bool("intersection", a, b, tr, rot)


def solid_bool(ope, a, b, tr, rot):
    s = Box()
    s[ope] = Box()
    s.name = f"{a.name}_{ope}_{b.name}"
    s[ope].a = a
    s[ope].b = b
    s[ope].translation = tr
    s[ope].rotation = rot
    return s


class BooleanVolume(gate.VolumeBase):
    """
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html?highlight=boolean#solids-made-by-boolean-operations
    """

    type_name = "Boolean"

    def __init__(self, name):
        gate.VolumeBase.__init__(self, name)
        # default values
        self.user_info.nodes = []
        # short
        self.user_info.add_node = (
            lambda x, y, z=Rotation.identity().as_matrix(): self.add_node(x, y, z)
        )
        # keep all created solids
        self.solid = self.user_info.solid  # None
        self.g4_solids = []

    def set_solid(self, solid):
        self.solid = solid

    def add_node(self, solid, translation=None, rotation_matrix=rid):
        if translation is None:
            translation = [0, 0, 0]
        b = Box()
        b.solid = solid
        b.translation = translation
        b.rotation = rotation_matrix
        self.user_info.nodes.append(b)

    def build_solid(self):
        return self._build_one_solid(self.solid)

    def _build_one_solid(self, solid):
        # could be simple, or starting with union etc
        if isinstance(solid, Box):
            for op in bool_operators:
                if op in solid:
                    return self._build_solid_bool(solid.name, op, solid[op])
        # build a 'fake' solid/volume to get the build_solid function
        # add the key 'i_am_a_solid' to avoid key checking
        solid.i_am_a_solid = True
        vol = gate.new_element(solid)
        return vol.build_solid()

    def _build_solid_bool(self, name, op, s):
        translation = gate.vec_np_as_g4(s.translation)
        rotation = gate.rot_np_as_g4(s.rotation)
        sa = self._build_one_solid(s.a)
        sb = self._build_one_solid(s.b)
        solid = None
        if op == "subtraction":
            solid = g4.G4SubtractionSolid(name, sa, sb, rotation, translation)
        if op == "union":
            solid = g4.G4UnionSolid(name, sa, sb, rotation, translation)
        if op == "intersection":
            solid = g4.G4IntersectionSolid(name, sa, sb, rotation, translation)
        if not solid:
            gate.fatal(f"Error in _build_solid_bool. Wrong operator ? {op}")
        self.g4_solids.append(sa)
        self.g4_solids.append(sb)
        self.g4_solids.append(solid)
        return solid



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

"""
Create a simulation object. The class is 'gate.Simulation'.
The single object that will contain all parameters of the
simulation is called 'sim' here.
"""
sim = gate.Simulation()

"""
Main global options.
The 'sim' object contains a structure called 'user_info' that gather all global options.
- For example here, the verbosity is set (verbosity means texts that are displayed during
the simulation run, mostly for debug)
- 'visu', if ON, display a windows with a QT view of the scene.
- random_engine and random_seed control the pseudo random engine. We recommend MersenneTwister.
  A seed can be specified, e.g. 123456, for reproducible simulation. Or you can use 'auto', an random seed
  will be generated.
"""
ui = sim.user_info
ui.verbose_level = gate.DEBUG
ui.running_verbose_level = gate.RUN
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_engine = "MersenneTwister"
ui.random_seed = "auto"
print(ui)

"""
Units. Get some default units from G4. To define a value with a unit, e.g. do:
x = 123 * cm
"""
m = gate.g4_units("m")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")

"""
Set the world size (like in the Gate macro). World is the only volume created by default.
It is described by a dict-like structure, accessible by sim.world.
The size is set here, as a 3D vector. Default material is G4_AIR.
"""
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]
world.material = "G4_AIR"

"""
A simple waterbox volume is created. It is inserted into the simulation with 'add_volume'.
This function return a dict-like structure (called 'waterbox' here) with various parameters
(size, position in the world, material). Note that, like in Geant4, the coordinate system
of all volumes is the one of the mother volume (here the world).
"""
waterbox = sim.add_volume("Box", "Waterbox")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

"""
The physic list by default is 'QGSP_BERT_EMV' (see Geant4 doc).
"""
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMV"
cuts = p.production_cuts
um = gate.g4_units("um")
cuts.world.gamma = 700 * um
cuts.world.electron = 700 * um
cuts.world.positron = 700 * um
cuts.world.proton = 700 * um

"""
Create a source, called 'Default'. The type of the source is 'Generic'.
Several parameters (particle, energy, direction etc) are available in the
dict-like structure.
"""
source = sim.add_source("Generic", "Default")
source.particle = "gamma"
source.energy.mono = 80 * keV
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.n = 200000

"""
Add a single scorer (called 'actor'), of type 'SimulationStatisticsActor'.
This simple scorer store the number or Run/Events/Track/Steps of the simulation.
We recommend to always add such actor.
The flag 'track_types_flag' gives more detailed results about the tracks (particle type)
""" ""
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

"""
Create G4 objects, like in conventional Geant4 simulation.
"""
sim.initialize()

"""
Start the simulation ! You can relax and drink coffee.
(The commented line indicates how to indicate to Geant4 to verbose during the simulation).
"""
# sim.apply_g4_command("/run/verbose 1")
sim.start()

"""
Now the simulation is terminated. The results is retrieved and can be displayed.
"""
stats = sim.get_actor("Stats")
print(stats)

# Comparison with gate simulation
# gate_test4_simulation_stats_actor
# Gate mac/main.mac
stats_ref = gate.read_stat_file(
    pathFile
    / ".."
    / "data"
    / "gate"
    / "gate_test004_simulation_stats_actor"
    / "output"
    / "stat.txt"
)
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.01)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import pathlib
import os

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.number_of_threads = 1
# some units
cm = gate.g4_units("cm")
m = gate.g4_units("m")
deg = gate.g4_units("deg")

# set the world size like in the Gate macro
sim.world.size = [1 * m, 1 * m, 1 * m]

# add a simple volume
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
waterbox.material = "G4_WATER"

# volume where to confine
stuff = sim.add_volume("Cons", "stuff")
stuff.mother = "waterbox"
stuff.rmin1 = 0
stuff.rmax1 = 0.5 * cm
stuff.rmin2 = 0
stuff.rmax2 = 0.5 * cm
stuff.dz = 2 * cm
stuff.dphi = 360 * deg
stuff.translation = [-5 * cm, 0 * cm, 0 * cm]
stuff.material = "G4_WATER"

# daughter volume
stuffi = sim.add_volume("Cons", "stuff_inside")
stuffi.mother = stuff.name
stuffi.rmin1 = 0
stuffi.rmax1 = 0.4 * cm
stuffi.rmin2 = 0
stuffi.rmax2 = 0.4 * cm
stuffi.dz = 2 * cm
stuffi.dphi = 360 * deg
stuffi.translation = [-0.1 * cm, 0 * cm, 0 * cm]
stuffi.material = "G4_AIR"

# useful units
MeV = gate.g4_units("MeV")
keV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
deg = gate.g4_units("deg")
mm = gate.g4_units("mm")

# activity
activity = 500000 * Bq
# activity = 50 * Bq

# test confined source
source = sim.add_source("Generic", "non_confined_src")
source.mother = "stuff"
source.particle = "gamma"
source.activity = activity / ui.number_of_threads
source.position.type = "box"
source.position.size = [5 * cm, 5 * cm, 5 * cm]
source.direction.type = "momentum"
source.direction.momentum = [-1, 0, 0]
source.energy.type = "mono"
source.energy.mono = 1 * MeV

# test confined source
"""
   the source is confined in the given volume ('stuff'), it means that
   all particles will be emitted only in this volume.
   The 'box' type is required to defined a larger volume that 'stuff'.
   It is done here by computing the bounding box
   Daughter volumes of 'stuff' do not count : no particle will be generated
   from 'stuff_inside'
"""
source = sim.add_source("Generic", "confined_src")
source.mother = "stuff"
source.particle = "gamma"
source.activity = activity / ui.number_of_threads
source.position.type = "box"
source.position.size = gate.get_volume_bounding_box_size(sim, source.mother)
print("Source size", source.position.size)
pMin, pMax = gate.get_volume_bounding_limits(sim, source.mother)
source.position.confine = "stuff"
source.direction.type = "momentum"
source.direction.momentum = [1, 0, 0]
source.energy.type = "mono"
source.energy.mono = 1 * MeV

# actors
stats = sim.add_actor("SimulationStatisticsActor", "Stats")

dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test010-2-edep.mhd"
# dose.output = paths.output_ref / 'test010-2-edep.mhd'
dose.mother = "waterbox"
dose.size = [100, 100, 100]
dose.spacing = [2 * mm, 1 * mm, 1 * mm]

# create G4 objects
sim.initialize()

# print after init
print(sim)
print("Simulation seed:", sim.actual_random_seed)

# start simulation
sim.start()

# get results
stats = sim.get_actor("Stats")
print(stats)
# stats.write(paths.output_ref / 'test010_confine_stats.txt')

# tests
stats_ref = gate.read_stat_file(paths.output_ref / "test010_confine_stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.10)
is_ok = is_ok and gate.assert_images(
    paths.output_ref / "test010-2-edep.mhd",
    paths.output / "test010-2-edep.mhd",
    stats,
    tolerance=59,
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
import gatetools as gt
import itk
import json
import opengate.contrib.pet_philips_vereos as pet_vereos
from box import Box


def create_pet_simulation(sim, param):
    # units
    m = gate.g4_units("m")
    cm3 = gate.g4_units("cm3")
    Bq = gate.g4_units("Bq")
    BqmL = Bq / cm3

    # main parameters
    ui = sim.user_info
    ui.check_volumes_overlap = True
    ui.number_of_threads = 1
    # ui.random_seed = 123456
    param.ac = param.activity_Bqml * BqmL / ui.number_of_threads
    if ui.visu:
        param.ac = 1 * BqmL
        ui.number_of_threads = 1

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # phantom ?
    if param.use_gaga:
        param.phantom_type = None
    if param.phantom_type == "analytic":
        add_analytical_phantom(sim, param)
    if param.phantom_type == "vox":
        add_voxelized_phantom(sim, param)

    # PET ?
    if param.use_pet:
        add_pet(sim, param)

    # physic list
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    p.apply_cuts = True
    cuts = p.production_cuts
    mm = gate.g4_units("mm")
    cuts.world.gamma = 1 * m
    cuts.world.positron = 1 * m
    cuts.world.electron = 1 * m
    if "iec" in cuts:
        cuts.iec.gamma = 0.1 * mm
        cuts.iec.positron = 0.1 * mm
        cuts.iec.electron = 0.1 * mm
    if "pet" in cuts:
        cuts.pet.gamma = 1 * mm
        cuts.pet.positron = 1 * mm
        cuts.pet.electron = 1 * mm

    # source ? FIXME
    if param.use_gaga:
        add_gaga_source(sim, param)
    else:
        if param.source_type == "analytic":
            add_analytical_source(sim, param)
        else:
            add_voxelized_source(sim, param)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True


def add_analytical_phantom(sim, param):
    print("Phantom: IEC analytical")
    gate_iec.add_phantom(sim)


def add_voxelized_phantom(sim, param):
    print("Phantom: IEC voxelized: ", param.iec_vox_mhd)
    iec = sim.add_volume("Image", "iec")
    gate_iec.create_material()
    iec.image = param.iec_vox_mhd
    iec.material = "G4_AIR"
    labels = json.loads(open(param.iec_vox_json).read())
    # labels are not material, we assign the material belows
    # all spheres are water ; central hole is lung ; shell are plastic
    iec.voxel_materials = []
    for l in labels:
        mat = "IEC_PLASTIC"
        if "capillary" in l:
            mat = "G4_WATER"
        if "cylinder_hole" in l:
            mat = "G4_LUNG_ICRP"
        if "interior" in l:
            mat = "G4_WATER"
        if "sphere" in l:
            mat = "G4_WATER"
        if "world" in l:
            mat = "G4_AIR"
        if "shell" in l:
            mat = "IEC_PLASTIC"
        m = [labels[l], labels[l] + 1, mat]
        iec.voxel_materials.append(m)


def add_gaga_source(sim, p):
    if p.source_type == "analytic":
        return add_gaga_source_analytic_condition(sim, p)
    if p.source_type == "vox":
        return add_gaga_source_vox_condition(sim, p)


def add_pet(sim, param):
    # add the pet volume
    pet_vereos.add_pet(sim, "pet")

    # hits collection
    hc = sim.add_actor("HitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    hc.mother = crystal.name
    hc.output = ""
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "TrackVolumeCopyNo",
        "PreStepUniqueVolumeID",
        "PostStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("HitsAdderActor", "Singles")
    sc.mother = crystal.name
    sc.input_hits_collection = "Hits"
    sc.policy = "EnergyWeightedCentroidPosition"
    sc.skip_attributes = ["PreStepUniqueVolumeID", "PostStepUniqueVolumeID"]
    sc.output = param.pet_output


def get_spheres_activity(sim, p):
    # compute spheres param
    spheres_diam = [10, 13, 17, 22, 28, 37]
    ac = p.ac
    spheres_activity_concentration = [ac, ac, ac, ac, ac, ac]

    # unit
    cm3 = gate.g4_units("cm3")
    Bq = gate.g4_units("Bq")
    BqmL = Bq / cm3

    spheres_centers, spheres_volumes = gate_iec.get_default_sphere_centers_and_volumes()
    spheres_activity_ratio = []
    spheres_activity = []
    for diam, ac, volume, center in zip(
        spheres_diam, spheres_activity_concentration, spheres_volumes, spheres_centers
    ):
        activity = ac * volume
        print(
            f"Sphere {diam}: {str(center):<30} {volume / cm3:7.3f} cm3 "
            f"{activity / Bq:7.0f} Bq  {ac / BqmL:7.1f} BqmL"
        )
        spheres_activity.append(activity)

    total_activity = sum(spheres_activity)
    print(f"Total activity {total_activity / Bq:.0f} Bq")
    for activity in spheres_activity:
        spheres_activity_ratio.append(activity / total_activity)
    print("Activity ratio ", spheres_activity_ratio, sum(spheres_activity_ratio))

    return spheres_activity, spheres_centers, spheres_activity_ratio, spheres_diam


def add_gaga_source_analytic_condition(sim, p):
    # compute spheres param
    (
        spheres_activity,
        spheres_centers,
        spheres_activity_ratio,
        spheres_diam,
    ) = get_spheres_activity(sim, p)
    spheres_radius = [x / 2.0 for x in spheres_diam]
    total_activity = sum(spheres_activity)

    def gen_cond(n):
        n_samples = gate_iec.get_n_samples_from_ratio(n, spheres_activity_ratio)
        # (it is very important to shuffle when several spheres to avoid time artifact)
        cond = gate_iec.generate_pos_spheres(
            spheres_centers, spheres_radius, n_samples, shuffle=True
        )
        return cond

    # GAN source
    cm = gate.g4_units("cm")
    mm = gate.g4_units("mm")
    keV = gate.g4_units("keV")
    gsource = sim.add_source("GAN", "gaga")
    gsource.particle = "gamma"
    # no phantom, we consider attached to the world at origin
    gsource.activity = total_activity
    gsource.pth_filename = p.gaga_pth
    gsource.position_keys = ["X1", "Y1", "Z1", "X2", "Y2", "Z2"]
    gsource.direction_keys = ["dX1", "dY1", "dZ1", "dX2", "dY2", "dZ2"]
    gsource.energy_key = ["E1", "E2"]
    gsource.time_key = ["t1", "t2"]
    # time is added to the simulation time
    gsource.time_relative = True
    gsource.weight_key = None
    # particle are move backward with 10 cm
    gsource.backward_distance = 10 * cm
    # if the kinetic E is below this threshold, we set it to 0
    gsource.energy_threshold = 0.1 * keV
    gsource.batch_size = 1e5
    gsource.verbose_generator = True
    # set the generator and the condition generator
    gen = gate.GANSourceConditionalPairsGenerator(gsource, 210 * mm, gen_cond)
    gsource.generator = gen


def add_gaga_source_vox_condition(sim, p):
    cm = gate.g4_units("cm")
    mm = gate.g4_units("mm")
    keV = gate.g4_units("keV")

    (
        spheres_activity,
        spheres_centers,
        spheres_activity_ratio,
        spheres_diam,
    ) = get_spheres_activity(sim, p)
    total_activity = sum(spheres_activity)

    # GAN source
    gsource = sim.add_source("GAN", "gaga")
    gsource.particle = "gamma"
    gsource.activity = total_activity
    gsource.pth_filename = p.gaga_pth
    gsource.position_keys = ["X1", "Y1", "Z1", "X2", "Y2", "Z2"]
    gsource.direction_keys = ["dX1", "dY1", "dZ1", "dX2", "dY2", "dZ2"]
    gsource.energy_key = ["E1", "E2"]
    gsource.time_key = ["t1", "t2"]
    # time is added to the simulation time
    gsource.time_relative = True
    gsource.weight_key = None
    # particle are move backward with 10 cm
    gsource.backward_distance = 10 * cm
    # if the kinetic E is below this threshold, we set it to 0
    gsource.energy_threshold = 0.1 * keV
    gsource.batch_size = 1e5
    gsource.verbose_generator = True

    # set the generator and the condition generator
    voxelized_cond_generator = gate.VoxelizedSourceConditionGenerator(p.source_vox_mhd)
    gen = gate.GANSourceConditionalPairsGenerator(
        gsource, 210 * mm, voxelized_cond_generator.generate_condition
    )
    gsource.generator = gen


def add_analytical_source(sim, p):
    print("Source: IEC analytical")
    if p.phantom_type == "vox":
        return add_analytical_source_with_vox_phantom(sim, p)
    gate_iec.add_spheres_sources(
        sim,
        "iec",
        "source",
        [10, 13, 17, 22, 28, 37],
        [p.ac, p.ac, p.ac, p.ac, p.ac, p.ac],  # in BqmL
        verbose=True,
    )
    sources = sim.source_manager.user_info_sources
    for source in sources.values():
        source.particle = "e+"
        source.energy.type = p.radionuclide
        source.direction.type = "iso"


def add_analytical_source_with_vox_phantom(sim, p):
    mm = gate.g4_units("mm")
    (
        spheres_activity,
        spheres_centers,
        spheres_activity_ratio,
        spheres_diam,
    ) = get_spheres_activity(sim, p)

    i = 0
    for s in spheres_diam:
        source = sim.add_source("Generic", f"source_{i}")
        source.particle = "e+"
        source.energy.type = p.radionuclide
        source.direction.type = "iso"
        source.activity = spheres_activity[i]
        source.position.type = "sphere"
        source.position.radius = s * mm / 2.0
        source.position.translation = spheres_centers[i]
        i += 1


def add_voxelized_source(sim, p):
    Bq = gate.g4_units("Bq")
    mm3 = gate.g4_units("mm3")
    cm3 = gate.g4_units("cm3")
    # compute volume to convert Bqml in Bq
    img = itk.imread(p.source_vox_mhd)
    stats = Box(gt.imageStatistics(img, None, False, 10))
    info = gate.get_info_from_image(img)
    vol = stats.sum * info.spacing[0] * info.spacing[1] * info.spacing[2] * mm3
    print(f"Volume source {vol} mm3")
    ac = p.activity_Bqml * vol / cm3
    print(f"Activity {ac} Bq")
    ac = ac * Bq

    # source
    source = sim.add_source("Voxels", "vox")
    source.mother = "iec"
    source.particle = "e+"
    source.energy.type = p.radionuclide
    source.activity = ac
    source.image = p.source_vox_mhd
    source.direction.type = "iso"
    if p.phantom_type == "vox":
        source.position.translation = gate.get_translation_between_images_center(
            p.iec_vox_mhd, p.source_vox_mhd
        )



import opengate as gate
import opengate_core as g4


class HitsEnergyWindowsActor(g4.GateHitsEnergyWindowsActor, gate.ActorBase):
    """
    Consider a list of hits and arrrange them according to energy intervals.
    Input: one HitsCollection
    Output: as many HitCollections as the number of energy windows
    """

    type_name = "HitsEnergyWindowsActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.attributes = []
        user_info.output = "EnergyWindows.root"
        user_info.input_hits_collection = "Hits"
        user_info.channels = []
        user_info.skip_attributes = []
        user_info.clear_every = 1e5

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateHitsEnergyWindowsActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)
        self.fStepFillNames = user_info.attributes

    def __del__(self):
        pass

    def __str__(self):
        s = f"HitsEnergyWindowsActor {self.user_info.name}"
        return s

    def StartSimulationAction(
        self,
    ):  # not needed, only if need to do something in python
        g4.GateHitsEnergyWindowsActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateHitsEnergyWindowsActor.EndSimulationAction(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec

paths = gate.get_default_test_paths(__file__, "")
paths.output_ref = paths.output_ref / "test040_ref"

# create the simulation
sim = gate.Simulation()

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
cm3 = gate.g4_units("cm3")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
BqmL = Bq / cm3
sec = gate.g4_units("second")
deg = gate.g4_units("deg")
kBq = 1000 * Bq
MBq = 1000 * kBq

# main parameters
ui = sim.user_info
ui.check_volumes_overlap = True
ui.number_of_threads = 1
ui.random_seed = 123456
ac = 5e3 * BqmL / ui.number_of_threads
ui.visu = False
colli_flag = not ui.visu
if ui.visu:
    ac = 1 * BqmL  # per mL
    ui.number_of_threads = 1

# world size
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]
world.material = "G4_AIR"

# iec phantom
iec_phantom = gate_iec.add_phantom(sim)

# add an artificial tungsten bar
vint = sim.get_volume_user_info("iec_interior")
print(vint)
t = sim.add_volume("Box", "tung")
t.mother = vint.name
t.size = [3 * cm, 8 * cm, 10 * cm]
t.translation = [-9 * cm, 5 * cm, 5 * cm]
t.material = "G4_CADMIUM_TUNGSTATE"
t.color = [0, 0, 1, 1]

# test phase space
phsp_sphere_surface = sim.add_volume("Sphere", "phase_space_sphere")
phsp_sphere_surface.rmin = 215 * mm
phsp_sphere_surface.rmax = 216 * mm
phsp_sphere_surface.color = [1, 1, 1, 1]
phsp_sphere_surface.material = "G4_AIR"

# physic list
sim.set_cut("world", "all", 1 * mm)

# source sphere
gate_iec.add_spheres_sources(
    sim,
    "iec",
    "source1",
    [10, 13, 17, 22, 28, 37],
    [ac * 6, ac * 5, ac * 4, ac * 3, ac * 2, ac],
    verbose=True,
)

# with acceptance angle (?) # FIXME
sources = sim.source_manager.user_info_sources
for source in sources.values():
    source.particle = "e+"
    source.energy.type = "Ga68"
    source.direction.type = "iso"

# background source 1:10 ratio with sphere
# bg = gate_iec.add_background_source(sim, 'iec', 'source_bg', ac / 10, verbose=True)

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.output = paths.output / "test040_ref_stats.txt"

# store phsp of exiting particles (gamma only)
phsp = sim.add_actor("PhaseSpaceActor", "phsp")
phsp.mother = phsp_sphere_surface.name
phsp.attributes = [
    "KineticEnergy",
    "PrePosition",
    "PreDirection",
    "GlobalTime",
    "TimeFromBeginOfEvent",
    "EventID",
    "EventPosition",
    "EventDirection",
    "EventKineticEnergy",
]
phsp.output = paths.output / "test040_ref_phsp.root"
phsp.store_absorbed_event = True
f = sim.add_filter("ParticleFilter", "f")
f.particle = "gamma"
phsp.filters.append(f)
f = sim.add_filter("KineticEnergyFilter", "f")
f.energy_min = 100 * keV
phsp.filters.append(f)

# go
sim.initialize()
sim.start()

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
stats = sim.get_actor("Stats")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "test040_ref_stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.01)

# 426760*2*0.8883814158496728 = 758251.3

phsp = sim.get_actor("phsp")
ref = 17299
ae = phsp.fNumberOfAbsorbedEvents
err = abs(ae - ref) / ref
tol = 0.01
is_ok = err < tol and is_ok
gate.print_test(is_ok, f"Number of absorbed events: {ae} vs {ref} = {err * 100:.2f}%")

# No other tests here for the moment, will be used by test040_gan
gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

"""
Tentative to run TWO simulations in the same script.

Seems not really possible from Geant4 point of view, I did no manage to delete and create another RunManager.

"""

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_engine = "MersenneTwister"
ui.random_seed = 123654

# set the world size like in the Gate macro
m = gate.g4_units("m")
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]

# add a simple waterbox volume
waterbox = sim.add_volume("Box", "Waterbox")
cm = gate.g4_units("cm")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

# default source for tests
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
source = sim.add_source("Generic", "Default")
source.particle = "gamma"
source.energy.mono = 80 * keV
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 20 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# sim.save('a.json')
# sim = Simulation.load('a.json')
# NOT POSSIBLE YET: cannot get result AND a new RunManager

# create G4 objects
sim.initialize()

# start simulation
# sim.apply_g4_command("/run/verbose 1")
sim.start()
stats = sim.get_actor("Stats")
print(stats)

print("del run manager")
del sim.g4_RunManager
print("del run manager")

sim = gate.Simulation()

## NOT POSSIBLE for the moment
# because (among other) G4UImanager cannot be recreated once killed.
# even with py::nodelete

# main options
ui = sim.user_info
ui.g4_verbose = True
ui.g4_verbose_level = 1
ui.visu = False
ui.random_engine = "MersenneTwister"
ui.random_seed = 123654

m = gate.g4_units("m")
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]

source = sim.add_source("Generic", "Default")
source.particle = "gamma"
source.energy.mono = 80 * keV
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 20 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

sim.initialize()
sim.start()
stats = sim.get_actor("Stats")
print(stats)

# print('del sim')
# ui = sim.g4_ui  # to avoid delete
# print(ui)
# del sim.g4_RunManager
# print(ui)
# del sim
# print(ui)
# del ui
# ui = None
# print(ui)

# print('new sim')
# sim2 = gate.Simulation(ui)

# print('before init')

# sim2.initialize()
# print('after init')
# sim2.start()
# stats = sim2.get_actor('Stats')
# print(stats)



import opengate as gate
import opengate_core as g4
from box import Box


class PhysicsUserInfo:
    """
    This class is a simple structure that contains all user general options of ths physics list.

    """

    def __init__(self, simulation):
        # keep pointer to ref
        self.simulation = simulation

        # physics list and decay
        self.physics_list_name = None
        self.enable_decay = False

        # options related to the cuts
        self.production_cuts = Box()
        self.energy_range_min = None
        self.energy_range_max = None
        self.apply_cuts = None

        # special case for EM parameters -> G4 object
        self.g4_em_parameters = g4.G4EmParameters.Instance()

    def __del__(self):
        pass

    def __str__(self):
        s = (
            f"{self.physics_list_name}"
            f"apply cuts : {self.apply_cuts}\n"
            f"prod cuts : {self.production_cuts}"
        )
        return s



import itk
import numpy as np
import os
import opengate as gate
import matplotlib.pyplot as plt
import colored
from box import Box, BoxList
import scipy
from scipy import optimize
from scipy import stats
import gatetools.phsp as phsp
import pathlib
import uproot
import sys

import matplotlib.pyplot as plt


def test_ok(is_ok=False):
    if is_ok:
        s = "Great, tests are ok."
        s = "\n" + colored.stylize(s, gate.color_ok)
        print(s)
        # sys.exit(0)
    else:
        s = "Error during the tests !"
        s = "\n" + colored.stylize(s, gate.color_error)
        print(s)
        sys.exit(-1)


def delete_run_manager_if_needed(sim):
    # if sys.platform == 'darwin':
    # if sys.version_info >= (3, 7):
    gate.warning(
        "WARNING, we need to delete G4RunManager, otherwise, GIL bug (seg fault)"
    )
    if sim.g4_RunManager:
        del sim.g4_RunManager
    print("RunManager deleted.")


def read_stat_file(filename):
    p = os.path.abspath(filename)
    f = open(p, "r")
    a = gate.UserInfo("Actor", "SimulationStatisticsActor", filename)
    stat = gate.SimulationStatisticsActor(a)
    # stat.counts = Box()
    read_track = False
    for line in f:
        if "NumberOfRun" in line:
            stat.counts.run_count = int(line[len("# NumberOfRun    =") :])
        if "NumberOfEvents" in line:
            stat.counts.event_count = int(line[len("# NumberOfEvents = ") :])
        if "NumberOfTracks" in line:
            stat.counts.track_count = int(line[len("# NumberOfTracks =") :])
        if "NumberOfSteps" in line:
            stat.counts.step_count = int(line[len("# NumberOfSteps  =") :])
        sec = gate.g4_units("s")
        if "ElapsedTimeWoInit" in line:
            stat.counts.duration = float(line[len("# ElapsedTimeWoInit     =") :]) * sec
        if read_track:
            w = line.split()
            name = w[1]
            value = w[3]
            stat.counts.track_types[name] = value
        if "Track types:" in line:
            read_track = True
            stat.user_info.track_types_flag = True
            stat.counts.track_types = {}
        if "Date" in line:
            stat.date = line[len("# Date                       =") :]
    return stat


def print_test(b, s):
    if b:
        print(s)
    else:
        color = gate.color_error
        print(colored.stylize(s, color))
    return b


def assert_stats(stat1, stat2, tolerance=0, is_ok=True):
    if stat2.counts.event_count != 0:
        event_d = stat1.counts.event_count / stat2.counts.event_count * 100 - 100
    else:
        event_d = 100
    if stat2.counts.track_count != 0:
        track_d = stat1.counts.track_count / stat2.counts.track_count * 100 - 100
    else:
        track_d = 100
    if stat2.counts.step_count != 0:
        step_d = stat1.counts.step_count / stat2.counts.step_count * 100 - 100
    else:
        step_d = 100
    if stat2.pps != 0:
        pps_d = stat1.pps / stat2.pps * 100 - 100
    else:
        pps_d = 100

    if stat2.tps != 0:
        tps_d = stat1.tps / stat2.tps * 100 - 100
    else:
        tps_d = 100

    if stat2.sps != 0:
        sps_d = stat1.sps / stat2.sps * 100 - 100
    else:
        sps_d = 100

    b = stat1.counts.run_count == stat2.counts.run_count
    is_ok = b and is_ok
    print_test(b, f"Runs:         {stat1.counts.run_count} {stat2.counts.run_count} ")

    b = abs(event_d) <= tolerance * 100
    is_ok = b and is_ok
    st = f"(tol = {tolerance * 100:.2f} %)"
    print_test(
        b,
        f"Events:       {stat1.counts.event_count} {stat2.counts.event_count} : {event_d:+.2f} %  {st}",
    )

    b = abs(track_d) <= tolerance * 100
    is_ok = b and is_ok
    print_test(
        b,
        f"Tracks:       {stat1.counts.track_count} {stat2.counts.track_count} : {track_d:+.2f} %  {st}",
    )

    b = abs(step_d) <= tolerance * 100
    is_ok = b and is_ok
    print_test(
        b,
        f"Steps:        {stat1.counts.step_count} {stat2.counts.step_count} : {step_d:+.2f} %  {st}",
    )

    print_test(
        True,
        f"PPS:          {stat1.pps:.1f} {stat2.pps:.1f} : "
        f"{pps_d:+.1f}%    speedup = x{(pps_d + 100) / 100:.1f}",
    )
    print_test(
        True,
        f"TPS:          {stat1.tps:.1f} {stat2.tps:.1f} : "
        f"{tps_d:+.1f}%    speedup = x{(tps_d + 100) / 100:.1f}",
    )
    print_test(
        True,
        f"SPS:          {stat1.sps:.1f} {stat2.sps:.1f} : "
        f"{sps_d:+.1f}%    speedup = x{(sps_d + 100) / 100:.1f}",
    )

    # particles types (Track)
    if stat1.user_info.track_types_flag and stat2.user_info.track_types_flag:
        for item in stat1.counts.track_types:
            v1 = stat1.counts.track_types[item]
            if item in stat2.counts.track_types:
                v2 = stat2.counts.track_types[item]
            else:
                print_test(b, f"Track {item:8}{v1} 0")
                continue
            v_d = float(v1) / float(v2) * 100 - 100
            # b = abs(v_d) <= tolerance * 100
            # is_ok = b and is_ok
            print_test(b, f"Track {item:8}{v1} {v2} : {v_d:+.1f}%")
        for item in stat2.counts.track_types:
            v2 = stat2.counts.track_types[item]
            if item not in stat1.counts.track_types:
                print_test(b, f"Track {item:8}0 {v2}")

    # consistency check
    if stat1.user_info.track_types_flag:
        n = 0
        for t in stat1.counts.track_types.values():
            n += int(t)
        b = n == stat1.counts.track_count
        print_test(b, f"Tracks      : {stat1.counts.track_types}")
        if "track_types" in stat2.counts:
            print_test(b, f"Tracks (ref): {stat2.counts.track_types}")
        print_test(b, f"Tracks vs track_types : {stat1.counts.track_count} {n}")
        is_ok = b and is_ok

    return is_ok


def plot_img_axis(ax, img, label, axis="z"):
    if axis == "y":
        return plot_img_y(ax, img, label)
    if axis == "x":
        return plot_img_x(ax, img, label)
    return plot_img_z(ax, img, label)


def plot_img_z(ax, img, label):
    # get data in np (warning Z and X inverted in np)
    data = itk.GetArrayViewFromImage(img)
    y = np.sum(data, 2)
    y = np.sum(y, 1)
    x = np.arange(len(y)) * img.GetSpacing()[2]
    ax.plot(x, y, label=label)
    ax.legend()


def plot_img_y(ax, img, label):
    # get data in np (warning Z and X inverted in np)
    data = itk.GetArrayViewFromImage(img)
    y = np.sum(data, 2)
    y = np.sum(y, 0)
    x = np.arange(len(y)) * img.GetSpacing()[2]
    ax.plot(x, y, label=label)
    ax.legend()


def plot_img_x(ax, img, label):
    # get data in np (warning Z and X inverted in np)
    data = itk.GetArrayViewFromImage(img)
    y = np.sum(data, 1)
    y = np.sum(y, 0)
    x = np.arange(len(y)) * img.GetSpacing()[2]
    ax.plot(x, y, label=label)
    ax.legend()


def assert_images(
    ref_filename1, filename2, stats=None, tolerance=0, ignore_value=0, axis="z"
):
    # read image and info (size, spacing etc)
    ref_filename1 = gate.check_filename_type(ref_filename1)
    filename2 = gate.check_filename_type(filename2)
    img1 = itk.imread(ref_filename1)
    img2 = itk.imread(filename2)
    info1 = gate.get_info_from_image(img1)
    info2 = gate.get_info_from_image(img2)

    # check img info
    is_ok = True
    if not np.all(info1.size == info2.size):
        print_test(False, f"Sizes are different {info1.size} vs {info2.size} ")
        is_ok = False
    if not np.allclose(info1.spacing, info2.spacing):
        print_test(False, f"Spacing are different {info1.spacing} vs {info2.spacing} ")
        is_ok = False
    if not np.allclose(info1.origin, info2.origin):
        print_test(False, f"Origin are different {info1.origin} vs {info2.origin} ")
        is_ok = False
    if not np.all(info1.dir == info2.dir):
        print_test(False, f"Directions are different {info1.dir} vs {info2.dir} ")
        is_ok = False
    print_test(is_ok, f"Images with same size/spacing/origin/dir ? {is_ok}")

    # check pixels contents, global stats
    data1 = itk.GetArrayViewFromImage(img1).ravel()
    data2 = itk.GetArrayViewFromImage(img2).ravel()

    print(
        f"Image1: {info1.size} {info1.spacing} {info1.origin} sum={np.sum(data1):.2f} {ref_filename1}"
    )
    print(
        f"Image2: {info2.size} {info2.spacing} {info2.origin} sum={np.sum(data2):.2f} {filename2}"
    )

    # do not consider pixels with a value of zero (data2 is the reference)
    d1 = data1[data2 != ignore_value]
    d2 = data2[data2 != ignore_value]

    # normalise by event
    if stats is not None:
        d1 = d1 / stats.counts.event_count
        d2 = d2 / stats.counts.event_count

    # normalize by sum of d1
    s = np.sum(d2)
    d1 = d1 / s
    d2 = d2 / s

    # sum of absolute difference (in %)
    sad = np.fabs(d1 - d2).sum() * 100
    is_ok = is_ok and sad < tolerance
    print_test(
        is_ok,
        f"Image diff computed on {len(data2 != 0)}/{len(data2.ravel())} \n"
        f"SAD (per event/total): {sad:.2f} % "
        f" (tolerance is {tolerance :.2f} %)",
    )

    # plot
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    plot_img_axis(ax, img1, "reference", axis)
    plot_img_axis(ax, img2, "test", axis)
    n = filename2.replace(".mhd", "_test.png")
    print("Save image test figure :", n)
    plt.savefig(n)

    return is_ok


def exponential_func(x, a, b):
    return a * np.exp(-b * x)


def Gauss(x, A, x0, sigma):
    return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))


def fit_exponential_decay(data, start, end):
    bin_heights, bin_borders = np.histogram(np.array(data), bins="auto", density=True)
    bin_widths = np.diff(bin_borders)
    bin_centers = bin_borders[:-1] + bin_widths / 2

    popt, pcov = scipy.optimize.curve_fit(exponential_func, bin_centers, bin_heights)
    xx = np.linspace(start, end, 100)
    yy = exponential_func(xx, *popt)
    hl = np.log(2) / popt[1]

    return hl, xx, yy


def get_new_key_name(key):
    # Correspondence between 1) gate root <-> opengate or 2) gate phsp <-> opengate
    # the third parameter is a scaling factor
    # the fourth is tolerance ?
    corres = [
        ["edep", "TotalEnergyDeposit", 1, 0.001],
        ["energy", "TotalEnergyDeposit", 1, 0.001],
        ["Ekine", "KineticEnergy", 1, 0.001],
        ["time", "GlobalTime", 1e-9, 0.02],
        ["posX", "PostPosition_X", 1, 0.9],
        ["posY", "PostPosition_Y", 1, 0.9],
        ["posZ", "PostPosition_Z", 1, 0.7],
        ["globalPosX", "PostPosition_X", 1, 0.7],
        ["globalPosY", "PostPosition_Y", 1, 0.7],
        ["globalPosZ", "PostPosition_Z", 1, 0.7],
        ["X", "PrePosition_X", 1, 0.8],
        ["Y", "PrePosition_Y", 1, 0.8],
        ["Z", "PrePosition_Z", 1, 0.8],
        ["dX", "PreDirection_X", 1, 0.01],
        ["dY", "PreDirection_Y", 1, 0.01],
        ["dZ", "PreDirection_Z", 1, 0.01],
        ["Weight", "Weight", 1, 0.01],
        ["trackID", "TrackID", 1, 0.05],
    ]
    for p in corres:
        if p[0] == key:
            return p[1], p[2], p[3]
    return None, None, None


def get_keys_correspondence(keys):
    keys1 = []
    keys2 = []
    scalings = []
    tols = []
    for k in keys:
        k2, s2, tol = gate.get_new_key_name(k)
        if k2:
            keys1.append(k)
            keys2.append(k2)
            scalings.append(s2)
            tols.append(tol)
    return keys1, keys2, scalings, tols


def rel_diff(a, b):
    return np.divide(a - b, a, out=np.zeros_like(a), where=a != 0) * 100


def rel_diff_range(a, b):
    r = np.max(a) - np.min(a)
    return np.divide(np.fabs(a - b), r, out=np.zeros_like(a), where=r != 0) * 100


def get_branch(tree, keys, key):
    """
    Return a branch whether it is a numpy or a uproot tree
    """
    try:
        index = keys.index(key)
        return tree[:, index]
    except:
        return tree[key]


"""
Previous trial with Two-sample Kolmogorov-Smirnov test
- works well for small samples size
- but not clear how to set "alpha" (tolerance) for large set like root tree

=> abort. REPLACE BY WASSERSTEIN DISTANCE

        Two-sample K–S test
        https://en.wikipedia.org/wiki/Kolmogorov%E2%80%93Smirnov_test
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kstest.html#scipy.stats.kstest

    Two-sample Kolmogorov-Smirnov test
    two-sided: The null hypothesis is that the two distributions are identical,
    F(x)=G(x) for all x; the alternative is that they are not identical.
    If the KS statistic is small or the p-value is high, then we cannot
    reject the null hypothesis in favor of the alternative.

    https://stackoverflow.com/questions/10884668/two-sample-kolmogorov-smirnov-test-in-python-scipy

    Results can be interpreted as following:
    - You can either compare the statistic value given by python to the KS-test
    critical value table according to your sample size. When statistic value is higher
    than the critical value, the two distributions are different.

    - Or you can compare the p-value to a level of significance a, usually a=0.05 or 0.01
    (you decide, the lower a is, the more significant). If p-value is lower than a, then it
    is very probable that the two distributions are different.

"""


def compare_branches(
    tree1,
    keys1,
    tree2,
    keys2,
    key1,
    key2,
    tol=0.8,
    scaling1=1,
    scaling2=1,
    ax=False,
    nb_bins=200,
):
    """
    Compare with Wasserstein distance
    Works well, but not easy to set the tolerance value.
    """
    # get branches
    b1 = get_branch(tree1, keys1, key1) * scaling1
    b2 = get_branch(tree2, keys2, key2) * scaling2

    return compare_branches_values(b1, b2, key1, key2, tol, ax, nb_bins)


def compare_branches_values(b1, b2, key1, key2, tol=0.8, ax=False, nb_bins=200):
    """
    Compare with Wasserstein distance
    Works well, but not easy to set the tolerance value.
    """

    # get ranges
    brange1 = np.max(b1) - np.min(b1)
    brange2 = np.max(b2) - np.min(b2)
    # mean
    m1 = np.mean(b1)
    m2 = np.mean(b2)
    n1 = np.size(b1)
    n2 = np.size(b2)

    # Earth mover distance (Wasserstein)
    wass = stats.wasserstein_distance(b1, b2)
    ok = wass < tol
    oks = "pass"
    if not ok:
        oks = "fail"
    s = (
        f"N: {n1:7} vs {n2:7} -> means {m1:6.2f} vs {m2:6.2f} -> ranges: {brange1:6.2f} vs {brange2:6.2f} "
        f" -> w:{wass:4.3f} vs {tol:4.3f}  \t {key1:<20} {key2:<20}  -> {oks} (tol {tol})"
    )
    print_test(ok, s)
    # figure ?
    if ax:
        nb_bins = nb_bins
        label = f" {key1} $\mu$={m1:.2f}"
        ax.hist(
            b1, nb_bins, density=True, histtype="stepfilled", alpha=0.5, label=label
        )
        label = f" {key2} $\mu$={m2:.2f}"
        ax.hist(
            b2, nb_bins, density=True, histtype="stepfilled", alpha=0.5, label=label
        )
        ax.set_ylabel("Counts")
        ax.legend()
    return ok


def compare_trees(
    tree1,
    allkeys1,
    tree2,
    allkeys2,
    keys1,
    keys2,
    tols,
    scalings1,
    scalings2,
    fig=False,
    nb_bins=200,
):
    if fig:
        nb_fig = len(keys1)
        nrow, ncol = phsp.fig_get_nb_row_col(nb_fig)
        f, ax = plt.subplots(nrow, ncol, figsize=(25, 10))
    is_ok = True
    n = 0
    print("Compare branches with Wasserstein distance")
    for i in range(len(keys1)):
        if fig:
            a = phsp.fig_get_sub_fig(ax, i)
            n += 1
        else:
            a = False
        is_ok = (
            compare_branches(
                tree1,
                allkeys1,
                tree2,
                allkeys2,
                keys1[i],
                keys2[i],
                tols[i],
                scalings1[i],
                scalings2[i],
                a,
                nb_bins=nb_bins,
            )
            and is_ok
        )
    if fig:
        phsp.fig_rm_empty_plot(nb_fig, n, ax)
    return is_ok


def get_default_test_paths(f, gate_folder=None):
    p = Box()
    p.current = pathlib.Path(f).parent.resolve()
    p.data = p.current / ".." / "data"
    if gate_folder:
        p.gate = p.current / ".." / "data" / "gate" / gate_folder
        p.gate_output = p.gate / "output"
        p.gate_data = p.gate / "data"
    p.output = p.current / ".." / "output"
    p.output_ref = p.current / ".." / "data" / "output_ref"
    return p


def compare_root2(root1, root2, branch1, branch2, keys, img_filename, n_tol=3):
    hits1 = uproot.open(root1)[branch1]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    hits2 = uproot.open(root2)[branch2]
    hits2_n = hits2.num_entries
    hits2 = hits2.arrays(library="numpy")

    print(f"Reference tree: {os.path.basename(root1)} n={hits1_n}")
    print(f"Current tree:   {os.path.basename(root2)} n={hits2_n}")
    diff = gate.rel_diff(float(hits1_n), float(hits2_n))
    is_ok = gate.print_test(
        np.fabs(diff) < n_tol,
        f"Difference: {hits1_n} {hits2_n} {diff:.2f}% (tol = {n_tol:.2f})",
    )
    print(f"Reference tree: {hits1.keys()}")
    print(f"Current tree:   {hits2.keys()}")

    keys = BoxList(keys)
    keys1 = [k.k1 for k in keys]
    keys2 = [k.k2 for k in keys]
    scalings = [k.scaling for k in keys]
    tols = [k.tol for k in keys]
    is_ok = (
        gate.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2.keys()),
            keys1,
            keys2,
            tols,
            [1] * len(scalings),
            scalings,
            True,
        )
        and is_ok
    )

    # figure
    plt.suptitle(
        f"Values: ref {os.path.basename(root1)} {os.path.basename(root2)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img_filename)
    print(f"Figure in {img_filename}")

    return is_ok


def compare_root(root1, root2, branch1, branch2, checked_keys, img):
    hits1 = uproot.open(root1)[branch1]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    hits2 = uproot.open(root2)[branch2]
    hits2_n = hits2.num_entries
    hits2 = hits2.arrays(library="numpy")

    print(f"Reference tree: {os.path.basename(root1)} n={hits1_n}")
    print(f"Current tree:   {os.path.basename(root2)} n={hits2_n}")
    diff = gate.rel_diff(float(hits1_n), float(hits2_n))
    is_ok = gate.print_test(
        np.fabs(diff) < 6, f"Difference: {hits1_n} {hits2_n} {diff:.2f}%"
    )
    print(f"Reference tree: {hits1.keys()}")
    print(f"Current tree:   {hits2.keys()}")

    keys1, keys2, scalings, tols = gate.get_keys_correspondence(checked_keys)
    is_ok = (
        gate.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2.keys()),
            keys1,
            keys2,
            tols,
            [1] * len(scalings),
            scalings,
            True,
        )
        and is_ok
    )

    # figure
    plt.suptitle(
        f"Values: ref {os.path.basename(root1)} {os.path.basename(root2)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img)
    print(f"Figure in {img}")

    return is_ok


def compare_root3(
    root1,
    root2,
    branch1,
    branch2,
    keys1,
    keys2,
    tols,
    scalings1,
    scalings2,
    img,
    hits_tol=6,
    nb_bins=200,
):
    hits1 = uproot.open(root1)[branch1]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    hits2 = uproot.open(root2)[branch2]
    hits2_n = hits2.num_entries
    hits2 = hits2.arrays(library="numpy")

    print(f"Reference tree: {os.path.basename(root1)} n={hits1_n}")
    print(f"Current tree:   {os.path.basename(root2)} n={hits2_n}")
    diff = gate.rel_diff(float(hits1_n), float(hits2_n))
    b = np.fabs(diff) < hits_tol
    is_ok = gate.print_test(b, f"Difference: {hits1_n} {hits2_n} {diff:.2f}%")
    print(f"Reference tree: {hits1.keys()}")
    print(f"Current tree:   {hits2.keys()}")

    # keys1, keys2, scalings, tols = Gate.get_keys_correspondence(checked_keys)
    is_ok = (
        gate.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2.keys()),
            keys1,
            keys2,
            tols,
            scalings1,
            scalings2,
            True,
            nb_bins=nb_bins,
        )
        and is_ok
    )

    # figure
    plt.suptitle(
        f"Values: ref {os.path.basename(root1)} {os.path.basename(root2)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img)
    print(f"Figure in {img}")

    return is_ok


# https://stackoverflow.com/questions/4527942/comparing-two-dictionaries-and-checking-how-many-key-value-pairs-are-equal
def dict_compare(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    shared_keys = d1_keys.intersection(d2_keys)
    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys
    modified = {o: (d1[o], d2[o]) for o in shared_keys if d1[o] != d2[o]}
    same = set(o for o in shared_keys if d1[o] == d2[o])
    return added, removed, modified, same


# Edit by Andreas and Martina
def write_gauss_param_to_file(
    outputdir, planePositionsV, saveFig=False, fNamePrefix="plane", fNameSuffix="a.mhd"
):
    # create output dir, if it doesn't exist
    if not os.path.isdir(outputdir):
        os.mkdir(outputdir)

    print("fNameSuffix", fNameSuffix)
    print("write mu and sigma file to dir: ")
    print(outputdir)

    # Extract gauss param along the two dim of each plane
    sigma_values = []
    mu_values = []
    for i in planePositionsV:

        filename = fNamePrefix + str(i) + fNameSuffix
        filepath = outputdir / filename

        # Get data from file
        data, spacing, shape = read_mhd(filepath)

        # Figure output is saved only if fig names are provided
        fig_name = None
        if saveFig:
            fig_name = str(outputdir) + "/Plane_" + str(i) + fNameSuffix + "_profile"

        # Get relevant gauss param
        sigma_x, mu_x, sigma_y, mu_y = get_gauss_param_xy(
            data, spacing, shape, filepath=fig_name, saveFig=saveFig
        )
        sigma_values.append([i, sigma_x, sigma_y])
        mu_values.append([i, mu_x, mu_y])

    np.savetxt(
        outputdir / "sigma_values.txt",
        sigma_values,
        header="Plane_nr sigma_x sigma_y",
        comments="",
    )
    np.savetxt(
        outputdir / "mu_values.txt", mu_values, header="Plane_nr mu_x mu_y", comments=""
    )

    return sigma_values, mu_values


def get_gauss_param_xy(data, spacing, shape, filepath=None, saveFig=False):
    # Parameters along x
    parameters_x, img_x, _ = extract_gauss_param_1D(
        data, shape[2], spacing[0], axis=1, createFig=saveFig
    )
    sigma_x = parameters_x[2]
    mu_x = parameters_x[1]

    # Parameters along y
    parameters_y, img_y, _ = extract_gauss_param_1D(
        data, shape[1], spacing[1], axis=2, createFig=saveFig
    )
    sigma_y = parameters_y[2]
    mu_y = parameters_y[1]

    # Save plots
    if filepath is not None:
        img_x.savefig(filepath + "_x.png")
        img_y.savefig(filepath + "_y.png")
        plt.close(img_x)
        plt.close(img_y)

    return sigma_x, mu_x, sigma_y, mu_y


def extract_gauss_param_1D(data, length, spacing, axis=1, createFig=False):
    poseVec = create_position_vector(length, spacing)
    dose = np.squeeze(np.sum(data, axis=axis))  # integrate dose along axis
    parameters, fit = gaussian_fit(poseVec, dose)

    fig = None
    if createFig:
        fig = plot_gauss_fit(poseVec, dose, fit, show=False)

    return parameters, fig, max(fit)


def plot_gauss_fit(positionVec, dose, fit, show=False):
    fig, a = plt.subplots()
    a.plot(positionVec, dose, "o", label="data")
    a.plot(positionVec, fit, "-", label="fit")
    a.set_xlabel("Depth [mm]")
    a.set_ylabel("Dose")
    if show:
        plt.show()

    return fig


def create_position_vector(length, spacing):
    # cretae position vector, with origin in the image plane's center
    width = length * spacing
    positionVec = np.arange(0, width, spacing) - width / 2 + spacing / 2

    return positionVec


def Gauss(x, A, x0, sigma):
    return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))


def gaussian_fit(positionVec, dose):
    # Fit data with Gaussian func
    mean = sum(positionVec * dose) / sum(dose)
    sigma = np.sqrt(sum(dose * (positionVec - mean) ** 2) / sum(dose))
    parameters, covariance = optimize.curve_fit(
        Gauss, positionVec, dose, p0=[max(dose), mean, sigma]
    )
    fit = Gauss(positionVec, parameters[0], parameters[1], parameters[2])

    return parameters, fit


def read_mhd(filename):
    img = itk.imread(str(filename))
    data = itk.GetArrayViewFromImage(img)
    spacing = img.GetSpacing()
    shape = data.shape
    return data, spacing, shape


def create_2D_Edep_colorMap(filepath, show=False):
    img = itk.imread(str(filepath))
    data = itk.GetArrayViewFromImage(img)

    fig = plt.figure(figsize=(20, 20))
    ax = fig.add_subplot(111)
    ax.set_title("colorMap")
    plt.imshow(data[0, :, :])
    ax.set_aspect("equal")
    plt.colorbar(orientation="vertical")
    if show:
        plt.show()

    return fig


def compareGaussParamFromFile(sigma, ref, rel_tol=0, abs_tol=0, verb=False):
    if rel_tol == 0 and abs_tol == 0:
        print("\033[91m Please provide non-zero tolerance\033[0m")

    with open(sigma, "r") as c1:
        lines1 = np.asarray(c1.readlines()[1:])

    with open(ref, "r") as c2:
        lines_ref = np.asarray(c2.readlines()[1:])

    is_ok = True

    for l, l_r in np.stack((lines1, lines_ref), axis=-1):
        sig_x = float(l.split(" ")[1])
        sig_y = float(l.split(" ")[2])
        plane = float(l.split(" ")[0])

        sig_x_r = float(l_r.split(" ")[1])
        sig_y_r = float(l_r.split(" ")[2])

        diff_x = abs(sig_x - sig_x_r)
        diff_y = abs(sig_y - sig_y_r)

        reldiff_x = (abs(sig_x - sig_x_r) / sig_x_r) * 100
        reldiff_y = (abs(sig_y - sig_y_r) / sig_y_r) * 100

        if verb:
            print(
                "Plane {0}: value x is {1}mm, value x ref is {2}mm ".format(
                    plane, round(sig_x, 2), round(sig_x_r, 2)
                )
            )
            print(
                "Plane {0}: value y is {1}mm, value y ref is {2}mm ".format(
                    plane, round(sig_y, 2), round(sig_y_r, 2)
                )
            )

        if diff_x > abs_tol and reldiff_x > rel_tol:
            print(
                "\033[91m Plane {0}:  rel difference along x is {1}%, threshold is {2}% \033[0m".format(
                    plane, round(reldiff_x, 2), round(rel_tol, 2)
                )
            )
            print(
                "\033[91m Plane {0}:  abs difference along x is {1}mm, threshold is {2}mm \033[0m".format(
                    plane, round(diff_x, 2), round(abs_tol, 2)
                )
            )
            is_ok = False
        else:
            print("Plane " + str(plane) + " along x is ok")

        if diff_y > abs_tol and reldiff_y > rel_tol:
            print(
                "\033[91m Plane {0}:  rel difference along y is {1}%, threshold is {2}% \033[0m".format(
                    plane, round(reldiff_y, 2), round(rel_tol, 2)
                )
            )
            print(
                "\033[91m Plane {0}:  abs difference along y is {1}mm, threshold is {2}mm \033[0m".format(
                    plane, round(diff_y, 2), round(abs_tol, 2)
                )
            )
            is_ok = False
        else:
            print("Plane " + str(plane) + " along y is ok")

    if is_ok:
        print("differences below threshold")
    else:
        print("\033[91m differences NOT OK \033[0m")

    return is_ok


def compareGaussParamArrays(paramTestV, paramRefV, rel_tol=0, abs_tol=0, verb=False):
    if rel_tol == 0 and abs_tol == 0:
        print("\033[91m Please provide non-zero tolerance\033[0m")

    is_ok = True

    for l in np.column_stack((paramTestV, paramRefV)):
        plane = l[0]
        pTest_x = l[1]
        pTest_y = l[2]
        pRef_x = l[4]
        pRef_y = l[5]

        diff_x = abs(pTest_x - pRef_x)
        diff_y = abs(pTest_y - pRef_y)

        reldiff_x = (abs(pTest_x - pRef_x) / pRef_x) * 100
        reldiff_y = (abs(pTest_y - pRef_y) / pRef_y) * 100

        if verb:
            print(
                "Plane {0}: value x is {1}mm, value x ref is {2}mm ".format(
                    plane, round(pTest_x, 2), round(pRef_x, 2)
                )
            )
            print(
                "Plane {0}: value y is {1}mm, value y ref is {2}mm ".format(
                    plane, round(pTest_y, 2), round(pRef_y, 2)
                )
            )

        if diff_x > abs_tol and reldiff_x > rel_tol:
            print(
                "\033[91m Plane {0}:  rel difference along x is {1}%, threshold is {2}% \033[0m".format(
                    plane, round(reldiff_x, 2), round(rel_tol, 2)
                )
            )
            print(
                "\033[91m Plane {0}:  abs difference along x is {1}mm, threshold is {2}mm \033[0m".format(
                    plane, round(diff_x, 2), round(abs_tol, 2)
                )
            )
            is_ok = False
        else:
            print("Plane " + str(plane) + " along x is ok")

        if diff_y > abs_tol and reldiff_y > rel_tol:
            print(
                "\033[91m Plane {0}:  rel difference along y is {1}%, threshold is {2}% \033[0m".format(
                    plane, round(reldiff_y, 2), round(rel_tol, 2)
                )
            )
            print(
                "\033[91m Plane {0}:  abs difference along y is {1}mm, threshold is {2}mm \033[0m".format(
                    plane, round(diff_y, 2), round(abs_tol, 2)
                )
            )
            is_ok = False
        else:
            print("Plane " + str(plane) + " along y is ok")

    if is_ok:
        print("differences below threshold")
    else:
        print("\033[91m differences NOT OK \033[0m")

    return is_ok


def test_weights(expected_ratio, mhd_1, mhd_2, thresh=0.1):
    img1 = itk.imread(str(mhd_1))
    img2 = itk.imread(str(mhd_2))
    data1 = itk.GetArrayViewFromImage(img1).ravel()
    data2 = itk.GetArrayViewFromImage(img2).ravel()

    sum1 = np.sum(data1)
    sum2 = np.sum(data2)
    ratio = sum2 / sum1

    print("\nSum energy dep for phantom 1: ", sum1)
    print("MSum energy dep for phantom 2: ", sum2)
    print("Ratio is: ", ratio)
    print("Expected ratio is: ", expected_ratio)

    is_ok = False
    if abs(ratio - expected_ratio) < thresh:
        is_ok = True
    else:
        print("\033[91m Ratio not as expected \033[0m")

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test037_pet_hits_singles_base import *

paths = gate.get_default_test_paths(__file__, "gate_test037_pet")

# create the simulation
sim = gate.Simulation()
create_pet_simulation(sim, paths)

# timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 0.00005 * sec]]

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results
stats = sim.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
p = paths.gate / "output_test1"
stats_ref = gate.read_stat_file(p / "stats1.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check root hits
hc = sim.get_actor_user_info("Hits")
f = p / "output1.root"
is_ok = check_root_hits(paths, 1, f, hc.output) and is_ok

# check root singles
sc = sim.get_actor_user_info("Singles")
is_ok = check_root_singles(paths, 1, f, sc.output) and is_ok

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test042_gauss_gps")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 123456

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
km = gate.g4_units("km")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq

# add a material database
sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

#  change world size
world = sim.world
world.size = [600 * cm, 500 * cm, 500 * cm]
world.material = "Vacuum"

# waterbox
phantom = sim.add_volume("Box", "phantom")
phantom.size = [10 * cm, 10 * cm, 10 * cm]
phantom.translation = [-5 * cm, 0, 0]
phantom.material = "G4_WATER"
phantom.color = [0, 0, 1, 1]

# daughter
phantom_y = sim.add_volume("Box", "phantom_y")
phantom_y.mother = phantom.name
phantom_y.size = [2 * mm, 10 * cm, 2 * mm]
phantom_y.translation = [49 * mm, 0, 0]
phantom_y.material = "G4_WATER"
phantom_y.color = [0, 0, 1, 1]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_INCLXX_EMZ"
sim.set_cut("world", "all", 1000 * km)
# FIXME need SetMaxStepSizeInRegion ActivateStepLimiter

# default source for tests
source = sim.add_source("Generic", "mysource")
source.energy.mono = 40 * MeV
source.particle = "proton"
source.position.type = "disc"  # pos = Beam, shape = circle + sigma
# rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
# source.position.radius = 8 * mm
source.position.sigma_x = 8 * mm
source.position.sigma_y = 8 * mm
source.direction.type = "momentum"
source.direction.momentum = [-1, 0, 0]
source.activity = 100 * kBq

# add dose actor
dose = sim.add_actor("DoseActor", "doseInXZ")
dose.output = paths.output / "test042-lateral_xz.mhd"
dose.mother = phantom.name
dose.size = [250, 1, 250]
dose.spacing = [0.4, 100, 0.4]
dose.hit_type = "random"

dose = sim.add_actor("DoseActor", "doseInXY")
dose.output = paths.output / "test042-lateral_xy.mhd"
dose.mother = phantom.name
dose.size = [250, 250, 1]
dose.spacing = [0.4, 0.4, 100]
dose.hit_type = "random"

dose = sim.add_actor("DoseActor", "doseInYZ")
dose.output = paths.output / "test042-lateral_yz.mhd"
dose.mother = phantom.name
dose.size = [1, 250, 250]
dose.spacing = [100, 0.4, 0.4]
dose.hit_type = "random"

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("stats")
print(stat)

dose = sim.get_actor("doseInXZ")
print(dose)

# ----------------------------------------------------------------------------------------------------------------
# tests
print()
gate.warning("Tests stats file")
stats_ref = gate.read_stat_file(paths.gate_output / "stats.txt")
is_ok = gate.assert_stats(stat, stats_ref, 0.14)

print()
gate.warning("Difference for EDEP XZ")
is_ok = (
    gate.assert_images(
        paths.gate_output / "lateral_xz_Protons_40MeV_sourceShapeGaussian-Edep.mhd",
        sim.get_actor("doseInXZ").user_info.output,
        stat,
        tolerance=10,
        ignore_value=0,
    )
    and is_ok
)

print()
gate.warning("Difference for EDEP XY")
is_ok = (
    gate.assert_images(
        paths.gate_output / "lateral_xy_Protons_40MeV_sourceShapeGaussian-Edep.mhd",
        sim.get_actor("doseInXY").user_info.output,
        stat,
        tolerance=10,
        ignore_value=0,
        axis="y",
    )
    and is_ok
)

print()
gate.warning("Difference for EDEP YZ")
is_ok = (
    gate.assert_images(
        paths.gate_output / "lateral_yz_Protons_40MeV_sourceShapeGaussian-Edep.mhd",
        sim.get_actor("doseInYZ").user_info.output,
        stat,
        tolerance=30,
        ignore_value=0,
        axis="y",
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
import opengate.contrib.phantom_nema_iec_body as gate_iec
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test029_volume_time_rotation")

# create the main simulation object
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1

# some basic units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
cm3 = gate.g4_units("cm3")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
deg = gate.g4_units("deg")
sec = gate.g4_units("second")
BqmL = Bq / cm3
kBq = 1000 * Bq

# world size
world = sim.world
world.size = [2 * m, 2 * m, 2 * m]
world.material = "G4_AIR"

# spect head
spect = gate_spect.add_ge_nm67_spect_head(
    sim, "spect", collimator_type=False, debug=False
)
initial_rot = Rotation.from_euler("X", 90, degrees=True)
t, rot = gate.get_transform_orbiting([0, 25 * cm, 0], "Z", 0)
rot = Rotation.from_matrix(rot)
spect.translation = t
spect.rotation = (rot * initial_rot).as_matrix()

# iec phantom
iec_phantom = gate_iec.add_phantom(sim)

# two sources (no background yet)
activity_concentration = 5000 * BqmL / ui.number_of_threads
ac = activity_concentration
sources = gate_iec.add_spheres_sources(
    sim,
    "iec",
    "iec_source",
    [10, 13, 17, 22, 28, 37],
    [ac, ac, ac, ac, ac, ac],
    verbose=True,
)

for s in sources:
    s.particle = "gamma"
    s.energy.type = "mono"
    s.energy.mono = 140 * keV
    s.direction.type = "iso"
    s.direction.type = "momentum"
    s.direction.momentum = [0, 1, 0]
    s.direction.acceptance_angle.volumes = ["spect"]
    s.direction.acceptance_angle.intersection_flag = True

sources = gate_iec.add_spheres_sources(
    sim,
    "iec",
    "iec_source2",
    [10, 13, 17, 22, 28, 37],
    [ac, ac, ac, ac, ac, ac],
    verbose=True,
)

for s in sources:
    s.particle = "gamma"
    s.energy.type = "mono"
    s.energy.mono = 140 * keV
    s.direction.type = "iso"
    s.direction.type = "momentum"
    s.direction.momentum = [1, 0, 0]
    s.direction.acceptance_angle.volumes = ["spect"]
    s.direction.acceptance_angle.intersection_flag = True

# physic list
sim.set_physics_list("G4EmStandardPhysics_option4")
sim.set_cut("world", "all", 10 * mm)
sim.set_cut("spect", "all", 1 * mm)

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.output = paths.output / "stats029.txt"

# hits collection
hc = sim.add_actor("HitsCollectionActor", "Hits")
hc.mother = "spect_crystal"
hc.output = ""  # No output
hc.attributes = [
    "PostPosition",
    "TotalEnergyDeposit",
    "PostStepUniqueVolumeID",
    "GlobalTime",
]

# singles collection
sc = sim.add_actor("HitsAdderActor", "Singles")
sc.mother = hc.mother
sc.input_hits_collection = "Hits"
sc.policy = "EnergyWeightedCentroidPosition"
sc.output = hc.output

# EnergyWindows
cc = sim.add_actor("HitsEnergyWindowsActor", "EnergyWindows")
cc.mother = hc.mother
cc.input_hits_collection = "Singles"
cc.channels = [
    {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
    {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
]
cc.output = hc.output

# projection
proj = sim.add_actor("HitsProjectionActor", "Projection")
proj.mother = hc.mother
proj.input_hits_collections = ["Singles", "scatter", "peak140"]
proj.spacing = [4.41806 * mm, 4.41806 * mm]
proj.size = [128, 128]
proj.origin_as_image_center = False
proj.output = paths.output / "proj029.mhd"

# motion of the spect, create also the run time interval
motion = sim.add_actor("MotionVolumeActor", "Move")
motion.mother = spect.name
motion.translations = []
motion.rotations = []
n = 9
sim.run_timing_intervals = []
start = -90
gantry_rotation = start
end = 1 * sec / n
initial_rot = Rotation.from_euler("X", 90, degrees=True)
for r in range(n):
    t, rot = gate.get_transform_orbiting([0, 30 * cm, 0], "Z", gantry_rotation)
    rot = Rotation.from_matrix(rot)
    rot = rot * initial_rot
    rot = gate.rot_np_as_g4(rot.as_matrix())
    motion.translations.append(t)
    motion.rotations.append(rot)
    sim.run_timing_intervals.append([start, end])
    gantry_rotation += 10
    start = end
    end += 1 * sec / n

print(f"Run intervals: {sim.run_timing_intervals}")

# check actor priority: the MotionVolumeActor must be first
l = [l for l in sim.actor_manager.user_info_actors.values()]
sorted_actors = sorted(l, key=lambda d: d.priority)
print(f"Actors order: ", [[l.name, l.priority] for l in sorted_actors])

# initialize & start
sim.initialize()
sim.start()

# WARNING when "angle_acceptance_volume" is enabled, it is a bit faster (+50%)
# but the result is not exactly the same as without. This is because, even
# if the initial particle is not in the direction of the spect system,
# it can scatter and still reach the detector.
# We don't have the collimator here (to faster the simulation), this is why the
# difference is not negligible.
# (the reference has been made without the "angle_acceptance_volume" option).

# -------------------------
gate.warning("Compare stats")
stats = gate.read_stat_file(paths.output / "stats029.txt")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "stats029.txt")
print(
    f"Number of steps was {stats.counts.step_count}, force to the same value (because of angle acceptance). "
)
stats.counts.step_count = stats_ref.counts.step_count  # force to id
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.01)
print(is_ok)

gate.warning("Compare images")
# read image and force change the offset to be similar to old Gate
is_ok = (
    gate.assert_images(
        paths.output_ref / "proj029.mhd",
        paths.output / "proj029.mhd",
        stats,
        tolerance=50,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)
print(is_ok)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec

paths = gate.get_default_test_paths(__file__, "")
paths.output_ref = paths.output_ref / "test040_ref"

# create the simulation
sim = gate.Simulation()

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
cm3 = gate.g4_units("cm3")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
BqmL = Bq / cm3
sec = gate.g4_units("second")
deg = gate.g4_units("deg")
kBq = 1000 * Bq
MBq = 1000 * kBq

# main parameters
ui = sim.user_info
ui.check_volumes_overlap = True
ui.number_of_threads = 1
ui.random_seed = 123456
ac = 15 * BqmL
ui.visu = False
if ui.visu:
    ac = 10 * BqmL  # per mL
    ui.number_of_threads = 1

# world size
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]
world.material = "G4_AIR"

# iec phantom
iec_phantom = gate_iec.add_phantom(sim)

# cylinder for phsp
sph_surface = sim.add_volume("Sphere", "phase_space_sphere")
sph_surface.rmin = 210 * mm
sph_surface.rmax = 211 * mm
sph_surface.color = [0, 1, 0, 1]
sph_surface.material = "G4_AIR"

# physic list
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
sim.set_cut("world", "all", 1 * mm)

# source sphere
gate_iec.add_spheres_sources(
    sim,
    "iec",
    "source1",
    [10, 13, 17, 22, 28, 37],
    [ac, ac, ac, ac, ac, ac],
    verbose=True,
)

sources = sim.source_manager.user_info_sources
for source in sources.values():
    source.particle = "e+"
    source.energy.type = "Ga68"

# background source 1:10 ratio with sphere
bg = gate_iec.add_background_source(sim, "iec", "source_bg", ac / 7, verbose=True)
bg.particle = "e+"
bg.energy.type = "Ga68"

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.output = paths.output / "test040_train_stats.txt"

# filter gamma only
f = sim.add_filter("ParticleFilter", "f")
f.particle = "gamma"

# phsp
phsp = sim.add_actor("PhaseSpaceActor", "phase_space")
phsp.mother = "phase_space_sphere"
# we use PrePosition because this is the first step in the volume
phsp.attributes = [
    "KineticEnergy",
    "PrePosition",
    "PreDirection",
    "TimeFromBeginOfEvent",
    # needed for gan_flag
    "EventID",
    "EventKineticEnergy",
    # for conditional :
    "EventPosition",
    "EventDirection",
]
phsp.output = paths.output / "test040_train.root"
phsp.store_absorbed_event = (
    True  # this option allow to store all events even if absorbed
)
phsp.filters.append(f)
print(phsp)
print(phsp.output)

# go
sim.initialize()
sim.start()

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
stats = sim.get_actor("Stats")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "test040_train_stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check phsp
print()
gate.warning(f"Check root")
p = sim.get_actor("phase_space")
print(f"Number of absorbed : {p.fNumberOfAbsorbedEvents}")
ref_file = paths.output_ref / "test040_train.root"
hc_file = phsp.output
checked_keys = [
    "TimeFromBeginOfEvent",
    "KineticEnergy",
    "PrePosition_X",
    "PrePosition_Y",
    "PrePosition_Z",
    "PreDirection_X",
    "PreDirection_Y",
    "PreDirection_Z",
    "EventKineticEnergy",
    "EventPosition_X",
    "EventPosition_Y",
    "EventPosition_Z",
    "EventDirection_X",
    "EventDirection_Y",
    "EventDirection_Z",
]
scalings = [1] * len(checked_keys)
# scalings[0] = 1e-9  # time in ns
tols = [1.0] * len(checked_keys)
tols[checked_keys.index("TimeFromBeginOfEvent")] = 0.007
tols[checked_keys.index("KineticEnergy")] = 0.002
tols[checked_keys.index("PrePosition_X")] = 1.6
tols[checked_keys.index("PrePosition_Y")] = 1.6
tols[checked_keys.index("PrePosition_Z")] = 1.6
tols[checked_keys.index("PreDirection_X")] = 0.01
tols[checked_keys.index("PreDirection_Y")] = 0.01
tols[checked_keys.index("PreDirection_Z")] = 0.01
tols[checked_keys.index("EventKineticEnergy")] = 0.02
tols[checked_keys.index("EventPosition_X")] = 1.8
tols[checked_keys.index("EventPosition_Y")] = 1.8
tols[checked_keys.index("EventPosition_Z")] = 1.8
tols[checked_keys.index("EventDirection_X")] = 0.02
tols[checked_keys.index("EventDirection_Y")] = 0.02
tols[checked_keys.index("EventDirection_Z")] = 0.02
print(scalings, tols)
is_ok = (
    gate.compare_root3(
        ref_file,
        hc_file,
        "phase_space",
        "phase_space",
        checked_keys,
        checked_keys,
        tols,
        scalings,
        scalings,
        paths.output / "test040_train_phsp.png",
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4

a = "titi"
s = g4.G4String(a)
print("G4String =", s, type(s))
assert s == "titi"
assert type(s) == gate.g4.G4String

t = g4.G4String(s)
print("G4String =", t, type(t))
assert t == "titi"

t = "toto"
print(f"s={s} t={t} ", type(t))

assert s == "titi"
assert t == "toto"

gate.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

paths = gate.get_default_test_paths(__file__, "gate_test027_fake_spect")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 2

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")

# world size
world = sim.world
world.size = [2 * m, 2 * m, 2 * m]

# material
sim.add_material_database(paths.data / "GateMaterials.db")

# fake spect head
waterbox = sim.add_volume("Box", "SPECThead")
waterbox.size = [55 * cm, 42 * cm, 18 * cm]
waterbox.material = "G4_AIR"

# crystal
crystal = sim.add_volume("Box", "crystal")
crystal.mother = "SPECThead"
crystal.size = [55 * cm, 42 * cm, 2 * cm]
crystal.translation = [0, 0, 4 * cm]
crystal.material = "NaITl"
crystal.color = [1, 1, 0, 1]

# colli
"""colli = sim.add_volume('Box', 'colli')
colli.mother = 'SPECThead'
colli.size = [55 * cm, 42 * cm, 6 * cm]
colli.material = 'Lead'
hole = sim.add_volume('Polyhedra', 'hole')
hole.mother = 'colli'
h = 5.8 * cm
hole.zplane = [-h / 2, h - h / 2]
hole.radius_outer = [0.15 * cm, 0.15 * cm, 0.15 * cm, 0.15 * cm, 0.15 * cm, 0.15 * cm]
hole.translation = None
hole.rotation = None

size = [77, 100, 1]
#size = [7, 10, 1]
tr = [7.01481 * mm, 4.05 * mm, 0]

## not correct position
start = [-(size[0] * tr[0]) / 2.0, -(size[1] * tr[1]) / 2.0, 0]
r1 = gate.repeat_array('colli1', start, size, tr)
start[0] += 3.50704 * mm
start[1] += 2.025 * mm
r2 = gate.repeat_array('colli2', start, size, tr)
hole.repeat = r1 + r2"""

# physic list
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
p.enable_decay = False
cuts = p.production_cuts
cuts.world.gamma = 0.01 * mm
cuts.world.electron = 0.01 * mm
cuts.world.positron = 1 * mm
cuts.world.proton = 1 * mm

# default source for tests
source = sim.add_source("Generic", "Default")
source.particle = "gamma"
source.energy.mono = 140.5 * keV
source.position.type = "sphere"
source.position.radius = 4 * cm
source.position.translation = [0, 0, -15 * cm]
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 5000 * Bq / ui.number_of_threads

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# hits collection
hc = sim.add_actor("HitsCollectionActor", "Hits")
hc.mother = crystal.name
hc.output = paths.output / "test027.root"
hc.attributes = [
    "KineticEnergy",
    "PostPosition",
    "PrePosition",
    "TotalEnergyDeposit",
    "GlobalTime",
    "TrackVolumeName",
    "TrackID",
    "PreStepUniqueVolumeID",
    "PostStepUniqueVolumeID",
    "TrackVolumeCopyNo",
    "TrackVolumeInstanceID",
]

# singles collection
sc = sim.add_actor("HitsAdderActor", "Singles")
sc.mother = crystal.name
sc.input_hits_collection = "Hits"
sc.policy = "EnergyWinnerPosition"
# sc.policy = 'EnergyWeightedCentroidPosition'
# same filename, there will be two branches in the file
sc.output = hc.output

sec = gate.g4_units("second")
ui.running_verbose_level = 2
sim.run_timing_intervals = [
    [0, 0.33 * sec],
    [0.33 * sec, 0.66 * sec],
    [0.66 * sec, 1 * sec],
]

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# stat
gate.warning("Compare stats")
stats = sim.get_actor("Stats")
print(stats)
print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
stats.counts.run_count = 1  # force to 1
stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.07)

# root compare HITS
print()
gate.warning("Compare HITS")
gate_file = paths.gate_output / "spect.root"
checked_keys = ["posX", "posY", "posZ", "edep", "time", "trackId"]
gate.compare_root(
    gate_file, hc.output, "Hits", "Hits", checked_keys, paths.output / "test027.png"
)

# Root compare SINGLES
print()
gate.warning("Compare SINGLES")
gate_file = paths.gate_output / "spect.root"
checked_keys = ["globalposX", "globalposY", "globalposZ", "energy"]
gate.compare_root(
    gate_file,
    sc.output,
    "Singles",
    "Singles",
    checked_keys,
    paths.output / "test027_singles.png",
)

# this is the end, my friend
gate.test_ok(is_ok)



import opengate as gate


class ActorBase(gate.UserElement):
    """
    Store user information about an actor
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        # user properties shared for all actors
        user_info.mother = gate.__world_name__
        user_info.filters = []
        user_info.priority = 100

    def __init__(self, user_info):
        # type_name MUST be defined in class that inherit from ActorBase
        super().__init__(user_info)

    def __del__(self):
        pass

    def initialize(self):
        # 'l' must be self to avoid being deleted
        self.l = []
        for f in self.user_info.filters:
            e = gate.new_element(f, self.simulation)
            e.Initialize(f.__dict__)
            self.l.append(e)
        # this is a copy (append cannot be used because fFilters is a std::vector)
        self.fFilters = self.l

    def __str__(self):
        s = f"str ActorBase {self.user_info.name} of type {self.user_info.type_name}"
        return s



from .GANSourceDefaultGenerator import GANSourceDefaultGenerator
import sys
import time
import opengate as gate


class GANSourceConditionalPairsGenerator(GANSourceDefaultGenerator):
    """
    Generate pairs of particles with a GAN, considering conditional vectors.

    The parameter sphere_radius is required : it is the radius of the sphere
    that surround the source during the training.

    """

    def __init__(self, user_info, sphere_radius, generate_condition_function):
        user_info.is_paired = True
        super().__init__(user_info)
        self.sphere_radius = sphere_radius
        self.generate_condition = generate_condition_function

    def generate_condition(self, n):
        gate.fatal(
            f'Error: to use GANSourceConditionalPairsGenerator,  you must provide a function "f" '
            f'that take a single int "n" as input and generate n condition samples. '
            f'This function "f" must be set with generator.generate_condition = f'
        )
        return None

    def generator(self, source):

        # get the info
        g = self.gan
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ...", end="")
            sys.stdout.flush()

        # generate cond
        cond = self.generate_condition(n)

        # generate samples (this is the most time-consuming part)
        fake = self.gaga.generate_samples2(
            g.params,
            g.G,
            g.D,
            n=n,
            batch_size=n,
            normalize=False,
            to_numpy=False,  # next step (from_tlor_to_pairs) is in torch, not numpy
            cond=cond,
            silence=True,
        )

        # parametrisation
        keys = g.params["keys_list"]
        if self.sphere_radius is None:
            gate.fatal(
                f'Error the option "sphere_radius" must be set for the source "{self.user_info.name}"'
            )
        params = {
            "keys_list": keys,
            "radius": self.sphere_radius,
            "ignore_directions": False,
        }
        fake = self.gaga.from_tlor_to_pairs(fake, params)
        keys = params["keys_output"]
        g.params["keys_output"] = keys

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # back from torch to numpy
        fake = fake.cpu().data.numpy()

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f" done in {end - start:0.1f} sec (GPU={g.params.current_gpu})")



from .MaterialBuilder import *


class MaterialDatabase:
    """
    Manage a list of Geant4 materials and elements.
    Read the data from a txt file (compatible with Gate database)

    """

    def __init__(self, filename, material_databases):
        self.filename = filename
        self.materials = {}
        self.elements = {}
        self.material_builders = {}
        self.element_builders = {}
        self.current_section = "None"
        self.material_databases = material_databases
        self.read_builders()

    def __del__(self):
        pass

    def read_builders(self):
        f = open(self.filename, "r")
        line = f.readline()
        while line:
            self.read_item(f, line)
            line = f.readline()

    def read_item(self, f, line):
        line = line.strip()
        if len(line) < 1:
            return
        if line[0] == "#":
            return
        # check if the current section change
        w = line.split()[0]
        if w == "[Elements]":
            self.current_section = "element"
            return
        if w == "[Materials]":
            self.current_section = "material"
            return
        if not self.current_section:
            gate.fatal(
                f"Error while reading the file {self.filename}, "
                f"current section is {self.current_section}. "
                f"File must start with [Elements] or [Materials]"
            )
        b = MaterialBuilder(self)
        if self.current_section == "element":
            b.read_element(f, line)
            self.element_builders[b.name] = b
        if self.current_section == "material":
            b.read_material(f, line)
            self.material_builders[b.name] = b

    def FindOrBuildMaterial(self, material):
        if material in self.materials:
            return self.materials[material]
        if material not in self.material_builders:
            return None
        b = self.material_builders[material]
        bm = b.build()
        self.materials[material] = bm
        return bm

    def FindOrBuildElement(self, element, loop=True):
        if element in self.elements:
            return self.elements[element]
        if element not in self.element_builders:
            if not loop:
                return None
            for n in self.material_databases:
                if n == self.filename:
                    continue
                db = self.material_databases[n]
                e = db.FindOrBuildElement(element, False)
                if e:
                    return e
            gate.fatal(f"Cannot find or build {element}")
            return None
        b = self.element_builders[element]
        be = b.build()
        self.elements[element] = be
        return be

    def dump_materials(self, level):
        if level == 0:
            return list(self.material_builders.keys())
        s = ""
        for m in self.material_builders.values():
            s = s + str(m) + "\n"
        return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.pet_philips_vereos as pet_vereos
import opengate.contrib.phantom_necr as phantom_necr


def create_pet_simulation(sim, paths, nb=1, actor_name="Singles"):
    """
    Simulation of a PET VEREOS with NEMA NECR phantom.
    - phantom is a simple cylinder and linear source
    - output is hits and singles only (no coincidences)
    - also digitizer is simplified: only raw hits and adder (for singles)
    """

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.check_volumes_overlap = False
    ui.random_seed = 123456789

    # units
    m = gate.g4_units("m")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    MBq = Bq * 1e6
    sec = gate.g4_units("second")

    #  change world size
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    # add a PET VEREOS
    sim.add_material_database(paths.gate_data / "GateMaterials_pet.db")
    pet = pet_vereos.add_pet(sim, "pet", create_housing=True, create_mat=False)

    # add table
    bed = pet_vereos.add_table(sim, "pet")

    # add NECR phantom
    phantom = phantom_necr.add_necr_phantom(sim, "phantom")

    # physics
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    sim.set_cut("world", "all", 1 * m)
    sim.set_cut(phantom.name, "all", 0.1 * mm)
    sim.set_cut(bed.name, "all", 0.1 * mm)
    sim.set_cut(f"{pet.name}_crystal", "all", 0.1 * mm)

    # default source for tests
    source = phantom_necr.add_necr_source(sim, phantom)
    total_yield = gate.get_rad_yield("F18")
    print("Yield for F18 (nb of e+ per decay) : ", total_yield)
    source.activity = 3000 * Bq * total_yield
    source.activity = 1787.914158 * MBq * total_yield
    source.half_life = 6586.26 * sec
    source.energy.type = "F18_analytic"  # WARNING not ok, but similar to previous Gate
    # source.energy.type = "F18"  # this is the correct F18 e+ source

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # hits collection
    hc = sim.add_actor("HitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    hc.mother = crystal.name
    print("Crystal :", crystal.name)
    hc.output = paths.output / f"test037_test{nb}.root"
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "TrackVolumeCopyNo",
        "PreStepUniqueVolumeID",
        "PostStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("HitsAdderActor", actor_name)
    sc.mother = crystal.name
    sc.input_hits_collection = "Hits"
    # sc.policy = "EnergyWinnerPosition"
    sc.policy = "EnergyWeightedCentroidPosition"
    sc.output = hc.output

    # timing
    sim.run_timing_intervals = [[0, 0.0001 * sec]]

    return sim


def default_root_hits_branches():
    k1 = ["posX", "posY", "posZ", "edep", "time"]
    k2 = [
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
        "TotalEnergyDeposit",
        "GlobalTime",
    ]
    return k1, k2


def default_root_singles_branches():
    k1 = ["globalPosX", "globalPosY", "globalPosZ", "energy", "time"]
    k2 = [
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
        "TotalEnergyDeposit",
        "GlobalTime",
    ]
    return k1, k2


def check_root_hits(paths, nb, ref_hits_output, hits_output):
    # check phsp (new version)
    print()
    gate.warning(f"Check root (hits)")
    k1, k2 = default_root_hits_branches()
    p1 = gate.root_compare_param_tree(ref_hits_output, "Hits", k1)
    # in the legacy gate, some edep=0 are still saved in the root file,
    # so we don't count that ones in the histogram comparison
    p1.mins[k1.index("edep")] = 0
    p2 = gate.root_compare_param_tree(hits_output, "Hits", k2)
    p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
    p = gate.root_compare_param(
        p1.the_keys, paths.output / f"test037_test{nb}_hits.png"
    )
    p.hits_tol = 6  # 5% tolerance (including the edep zeros)
    p.tols[k1.index("posX")] = 6
    p.tols[k1.index("posY")] = 6
    p.tols[k1.index("posZ")] = 1.5
    p.tols[k1.index("edep")] = 0.002
    p.tols[k1.index("time")] = 0.0001
    is_ok = gate.root_compare4(p1, p2, p)

    return is_ok


def check_root_singles(paths, nb, ref_singles_output, singles_output):
    # check phsp (singles)
    print()
    gate.warning(f"Check root (singles)")
    k1, k2 = default_root_singles_branches()
    p1 = gate.root_compare_param_tree(ref_singles_output, "Singles", k1)
    # in the legacy gate, some edep=0 are still saved in the root file,
    # so we don't count that ones in the histogram comparison
    p1.mins[k1.index("energy")] = 0
    p2 = gate.root_compare_param_tree(singles_output, "Singles", k2)
    p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
    p = gate.root_compare_param(
        p1.the_keys, paths.output / f"test037_test{nb}_singles.png"
    )
    p.hits_tol = 5  # 5% tolerance (including the edep zeros)
    p.tols[k1.index("globalPosX")] = 3
    p.tols[k1.index("globalPosY")] = 3
    p.tols[k1.index("globalPosZ")] = 1.5
    p.tols[k1.index("energy")] = 0.004
    p.tols[k1.index("time")] = 0.0001

    is_ok = gate.root_compare4(p1, p2, p)

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test037_pet_hits_singles_base import *

paths = gate.get_default_test_paths(__file__, "gate_test037_pet")

v = "2_2"

# create the simulation
sim = gate.Simulation()
create_pet_simulation(sim, paths, v, "Singles_adder")


hc = sim.get_actor_user_info("Hits")

# change the digitizer adder
sc = sim.get_actor_user_info("Singles_adder")
sc.policy = "EnergyWeightedCentroidPosition"
# sc.output = None # do not save it
# gate.print_dic(sc)
import json

s = json.dumps(sc.__dict__, indent=2, default=str)
print(s)

# singles collection
module = sim.get_volume_user_info("pet_module")
die = sim.get_volume_user_info("pet_die")
sc = sim.add_actor("HitsAdderActor", "Singles")
sc.mother = module.name
sc.input_hits_collection = "Singles_adder"
# sc.policy = "EnergyWinnerPosition"
sc.policy = "EnergyWeightedCentroidPosition"
sc.output = hc.output
# gate.print_dic(sc)

# timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 0.00003 * sec]]

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results
stats = sim.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
# p = paths.gate / "output_test1"
p = paths.gate / "output"
stats_ref = gate.read_stat_file(p / f"stats{v}.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check root hits
hc = sim.get_actor_user_info("Hits")
is_ok = check_root_hits(paths, "output", v, hc.output) and is_ok

# check root singles
sc = sim.get_actor_user_info("Singles")
is_ok = check_root_singles(paths, "output", v, sc.output) and is_ok

gate.test_ok(is_ok)
